# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Union
from urllib.parse import urlparse

from fastapi import WebSocket, status
from fastapi.exceptions import HTTPException
from httpx import AsyncClient
from structlog import get_logger

from orchestrator.security import oidc_user, opa_security_default
from orchestrator.websocket.managers.broadcast_websocket_manager import BroadcastWebsocketManager
from orchestrator.websocket.managers.memory_websocket_manager import MemoryWebsocketManager

logger = get_logger(__name__)


class WebSocketManager:
    _backend: Union[MemoryWebsocketManager, BroadcastWebsocketManager]

    def __init__(self, websockets_enabled: bool, broadcast_url: str):
        self.enabled = websockets_enabled
        self.broadcaster_type = urlparse(broadcast_url).scheme
        self.connected = False
        if self.broadcaster_type == "redis":
            self._backend = BroadcastWebsocketManager(broadcast_url)
        else:
            self._backend = MemoryWebsocketManager()

    async def authorize(self, websocket: WebSocket, token: str) -> Optional[Dict]:
        try:
            async with AsyncClient() as client:
                user = await oidc_user(websocket, token=token)  # type: ignore
                if user:
                    await opa_security_default(websocket, user, client)  # type: ignore
        except HTTPException as e:
            return {"error": vars(e)}
        return None

    async def connect_redis(self) -> None:
        if not self.connected:
            await self._backend.connect_redis()
            self.connected = True

    async def disconnect_redis(self) -> None:
        if self.connected:
            await self._backend.disconnect_redis()
            self.connected = False

    async def connect(self, websocket: WebSocket, channel: str) -> None:
        await self._backend.connect(websocket, channel)

    async def disconnect(
        self, websocket: WebSocket, code: int = status.WS_1000_NORMAL_CLOSURE, reason: Union[Dict, str, None] = None
    ) -> None:
        await self._backend.disconnect(websocket, code, reason)

    async def disconnect_all(self) -> None:
        await self._backend.disconnect_all()

    async def broadcast_data(self, channels: List[str], data: Dict) -> None:
        await self._backend.broadcast_data(channels, data)



from http import HTTPStatus

from orchestrator.db import EngineSettingsTable, db


def test_get_engine_status(test_client):
    engine_settings = EngineSettingsTable.query.one()
    response = test_client.get("/api/settings/status")
    assert response.status_code == HTTPStatus.OK
    assert response.json()["running_processes"] == 0
    assert response.json()["global_lock"] is False
    assert response.json()["global_status"] == "RUNNING"

    engine_settings.global_lock = True
    engine_settings.running_processes = 1
    db.session.flush()

    response = test_client.get("/api/settings/status")
    assert response.status_code == HTTPStatus.OK
    assert response.json()["running_processes"] == 1
    assert response.json()["global_lock"] is True
    assert response.json()["global_status"] == "PAUSING"

    engine_settings.running_processes = 0
    db.session.flush()

    response = test_client.get("/api/settings/status")
    assert response.status_code == HTTPStatus.OK
    assert response.json()["running_processes"] == 0
    assert response.json()["global_lock"] is True
    assert response.json()["global_status"] == "PAUSED"


def test_set_engine_status(test_client):
    response = test_client.put("/api/settings/status", json={"global_lock": True})
    assert response.status_code == HTTPStatus.OK
    assert response.json()["running_processes"] == 0
    assert response.json()["global_lock"] is True
    assert response.json()["global_status"] == "PAUSED"

    response = test_client.put("/api/settings/status", json={"global_lock": False})
    assert response.status_code == HTTPStatus.OK
    assert response.json()["running_processes"] == 0
    assert response.json()["global_lock"] is False
    assert response.json()["global_status"] == "RUNNING"



from typing import Dict, List, Set, Union

from more_itertools import flatten
from sqlalchemy.sql.expression import Delete, Insert, Update
from sqlalchemy.sql.selectable import ScalarSelect

from orchestrator.cli.domain_gen_helpers.helpers import sql_compile
from orchestrator.cli.domain_gen_helpers.product_helpers import get_product_id, get_product_ids
from orchestrator.cli.helpers.input_helpers import get_user_input
from orchestrator.cli.helpers.print_helpers import COLOR, print_fmt, str_fmt
from orchestrator.db.models import FixedInputTable


def map_update_fixed_inputs(product_diffs: Dict[str, Dict[str, Set[str]]]) -> Dict[str, Dict[str, str]]:
    """Map fixed inputs to update.

    Args:
        - product_diffs: Dict with product differences.
            - key: product name
            - value: Dict with differences between model and database.
                - key: difference name, 'missing_fixed_inputs_in_model' and 'missing_fixed_inputs_in_db' are used to check if a fixed input can be renamed.
                - value: Set of fixed input names.

    Returns: Dict with updated fixed inputs mapped by product.
        - key: product name.
        - value: Dict with updated fixed inputs.
            - key: old fixed input name.
            - value: new fixed input name.
    """
    print_fmt("\nUpdate fixed inputs", flags=[COLOR.BOLD, COLOR.UNDERLINE])

    def rename_map(product_name: str, product_diff: Dict[str, Set[str]]) -> Dict[str, str]:
        db_props = list(product_diff.get("missing_fixed_inputs_in_model", []))
        model_props = list(product_diff.get("missing_fixed_inputs_in_db", []))

        # If 1 field differs between model and db, ask if this is a renaming, otherwise delete/create all differing fields
        if len(db_props) == 1 and len(model_props) == 1:
            should_rename = get_user_input(
                "".join(
                    [
                        "Do you wish to rename fixed input ",
                        str_fmt(db_props[0], flags=[COLOR.MAGENTA]),
                        " to ",
                        str_fmt(model_props[0], flags=[COLOR.MAGENTA]),
                        " for product ",
                        str_fmt(product_name, flags=[COLOR.BOLD]),
                        "? [y/N]: ",
                    ]
                ),
                "N",
            )
            if should_rename == "y":
                product_diffs[product_name]["missing_fixed_inputs_in_model"] = set()
                product_diffs[product_name]["missing_fixed_inputs_in_db"] = set()
                return {db_props[0]: model_props[0]}
        return {}

    updates = {name: rename_map(name, diff) for name, diff in product_diffs.items()}
    return {k: v for k, v in updates.items() if v}


def generate_create_fixed_inputs_sql(
    fixed_inputs: Dict[str, Set[str]], inputs: Dict[str, Dict[str, str]], revert: bool = False
) -> List[str]:
    """Generate SQL to create fixed inputs.

    Args:
        - fixed_inputs: Dict with product names by fixed input.
            - key: fixed input value.
            - value: product names.
        - inputs: Optional Dict to prefill fixed input 'value' per product.
            - key: fixed input name.
            - value: Dict with prefilled value.
                - key: 'value' as key.
                - value: prefilled value.
        - revert: boolean to create SQL string with value filled by the database.

    Returns: List of SQL to create fixed inputs.
    """
    print_fmt("\nCreate fixed inputs", flags=[COLOR.BOLD, COLOR.UNDERLINE])

    def create_fixed_input(fixed_input: str, product_names: Set[str]) -> str:
        def create_product_insert_dict(product_name: str) -> Dict[str, Union[str, ScalarSelect]]:
            product_id_sql = get_product_id(product_name)

            if revert:
                value = (
                    FixedInputTable.query.where(
                        FixedInputTable.name == fixed_input, FixedInputTable.product_id == product_id_sql
                    )
                    .with_entities(FixedInputTable.value)
                    .one()
                )[0]
            else:
                prompt = "".join(
                    [
                        "Supply fixed input value for product ",
                        str_fmt(product_name, flags=[COLOR.BOLD]),
                        " and fixed input ",
                        str_fmt(fixed_input, flags=[COLOR.MAGENTA]),
                        ": ",
                    ]
                )
                value = inputs.get(product_name, {}).get(fixed_input) or get_user_input(prompt)

            return {"name": fixed_input, "value": value, "product_id": product_id_sql}

        fixed_input_dicts = [create_product_insert_dict(product_name) for product_name in product_names]
        return str(sql_compile(Insert(FixedInputTable).values(fixed_input_dicts)))

    return [create_fixed_input(*item) for item in fixed_inputs.items()]


def generate_delete_fixed_inputs_sql(fixed_inputs: Dict[str, Set[str]]) -> List[str]:
    """Generate SQL to delete fixed inputs.

    Args:
        - fixed_inputs: Dict with product names by fixed input.
            - key: fixed input value.
            - value: product names.

    Returns: List of SQL strings to delete fixed inputs.
    """

    def delete_fixed_input(fixed_input: str, product_names: Set[str]) -> str:
        product_ids_sql = get_product_ids(product_names)
        return str(
            sql_compile(
                Delete(FixedInputTable).where(
                    FixedInputTable.product_id.in_(product_ids_sql),
                    FixedInputTable.name == fixed_input,
                )
            )
        )

    return [delete_fixed_input(*item) for item in fixed_inputs.items()]


def generate_update_fixed_inputs_sql(product_fixed_inputs: Dict[str, Dict[str, str]]) -> List[str]:
    """Generate SQL to update fixed inputs.

    Args:
        - product_fixed_inputs: Dict with product names by fixed input.
            - key: fixed input value.
            - value: product names.

    Returns: List of SQL strings to update fixed inputs.
    """

    def update_fixed_inputs(product_name: str, fixed_inputs: Dict[str, str]) -> List[str]:
        product_id_sql = get_product_id(product_name)

        def update_fixed_input(old_name: str, new_name: str) -> str:
            return sql_compile(
                Update(FixedInputTable)
                .where(
                    FixedInputTable.product_id == (product_id_sql),
                    FixedInputTable.name == old_name,
                )
                .values(name=new_name)
            )

        return [update_fixed_input(*fixed_input) for fixed_input in fixed_inputs.items()]

    return list(flatten([update_fixed_inputs(*item) for item in product_fixed_inputs.items()]))



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import floor, log
from typing import Union


def format_bandwidth(mbits: int, short: bool = False) -> str:
    """Format bandwidth with unit designator (eg, Mbit/s or M).

    * It supports units up to and including Petabit.
    * If it cannot convert the number to an integral one, it will allow for 1 decimal.
    * Negative bandwidths will always return 0 Mbit/s or 0 M

    Examples::

        >>> format_bandwidth(40)
        '40 Mbit/s'
        >>> format_bandwidth(40, True)
        '40M'
        >>> format_bandwidth(10000)
        '10 Gbit/s'
        >>> format_bandwidth(10000, True)
        '10G'
        >>> format_bandwidth(1300)
        '1.3 Gbit/s'
        >>> format_bandwidth(0)
        '0 Mbit/s'
        >>> format_bandwidth(-100)
        '0 Mbit/s'

    Args:
        mbits: number of mbits
        short: boolean indicating whether to use 'M', 'G' or 'Mbit/s' or 'Gbit/s', etc

    Returns:
        Formatted number with unit designator.

    """
    unit_fmt, separator = ("short", "") if short else ("long", " ")
    units = {"short": ["M", "G", "T", "P"], "long": ["Mbit/s", "Gbit/s", "Tbit/s", "Pbit/s"]}
    base = 1_000
    magnitude = 0
    number = "0"
    if mbits > 0:
        magnitude = floor(log(mbits, base))
        number = f"{(mbits / base ** magnitude):.1f}"
    if number.endswith(".0"):
        number = number[:-2]
    return f"{number}{separator}{units[unit_fmt][magnitude]}"


def speed_humanize(bandwidth: Union[int, str], short: bool = False) -> str:
    if isinstance(bandwidth, str) and "," in bandwidth:
        return ",".join(map(speed_humanize, bandwidth.split(",")))
    try:
        speed = int(bandwidth)
    except ValueError:
        return str(bandwidth)
    except TypeError:
        return str(bandwidth)
    return format_bandwidth(speed, short=short)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Any, Dict

import structlog
from fastapi.routing import APIRouter
from more_itertools import first_true, one
from sqlalchemy.exc import NoResultFound

from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.schemas import FixedInputConfigurationSchema
from orchestrator.services import products

router = APIRouter()

logger = structlog.get_logger(__name__)


@router.get("/configuration", response_model=FixedInputConfigurationSchema)
def fi_configuration() -> Dict[str, Any]:
    product_tags = products.get_tags()

    data: dict = {"fixed_inputs": [], "by_tag": {}}
    for tag in product_tags:
        data["by_tag"][tag] = []

    for product_name, model in SUBSCRIPTION_MODEL_REGISTRY.items():
        try:
            product = products.get_product_by_name(product_name)
        except NoResultFound:
            logger.error(
                "Couldn't resolve product with fixed inputs for a domain model, due to a product_name " "mismatch.",
                product_name=product_name,
            )

        for fi_name, fi_type in model._non_product_block_fields_.items():
            fi_data = first_true(data["fixed_inputs"], None, lambda i: i["name"] == fi_name)  # noqa: B023
            if not fi_data:
                if issubclass(fi_type, Enum):
                    data["fixed_inputs"].append(
                        {
                            "name": fi_name,
                            "description": (fi_type.__doc__ or fi_name).splitlines()[0],
                            "values": list(map(lambda v: str(v.value), fi_type)),
                        }
                    )
                else:
                    raise ValueError(f"{fi_name} of {product_name} should be an enum.")

            if not first_true(data["by_tag"][product.tag], None, lambda i: fi_name in i):  # noqa: B023
                data["by_tag"][product.tag].append({fi_name: True})

    # Check if required
    for product_name, model in SUBSCRIPTION_MODEL_REGISTRY.items():
        product = products.get_product_by_name(product_name)

        for fi in data["by_tag"][product.tag]:
            if one(fi) not in model._non_product_block_fields_:
                fi[one(fi)] = False

    return data



from typing import Dict, Set, Type

from pydantic import BaseModel

from orchestrator.domain.base import ProductBlockModel, SubscriptionModel


class DomainModelChanges(BaseModel):
    create_products: Dict[str, Type[SubscriptionModel]] = {}
    delete_products: Set[str] = set()
    create_product_to_block_relations: Dict[str, Set[str]] = {}
    delete_product_to_block_relations: Dict[str, Set[str]] = {}
    create_product_blocks: Dict[str, Type[ProductBlockModel]] = {}
    delete_product_blocks: Set[str] = set()
    create_product_block_relations: Dict[str, Set[str]] = {}
    delete_product_block_relations: Dict[str, Set[str]] = {}
    create_product_fixed_inputs: Dict[str, Set[str]] = {}
    update_product_fixed_inputs: Dict[str, Dict[str, str]] = {}
    delete_product_fixed_inputs: Dict[str, Set[str]] = {}
    create_resource_types: Set[str] = set()
    rename_resource_types: Dict[str, str] = {}
    update_block_resource_types: Dict[str, Dict[str, str]] = {}
    delete_resource_types: Set[str] = set()
    create_resource_type_relations: Dict[str, Set[str]] = {}
    create_resource_type_instance_relations: Dict[str, Set[str]] = {}
    delete_resource_type_relations: Dict[str, Set[str]] = {}


class DuplicateException(Exception):
    pass


class ModelUpdates(BaseModel):
    fixed_inputs: Dict[str, Dict[str, str]] = {}
    resource_types: Dict[str, str] = {}
    block_resource_types: Dict[str, Dict[str, str]] = {}



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os import getenv
from typing import Any, Dict, Optional, Tuple
from uuid import UUID

from redis import Redis
from redis.asyncio import Redis as AIORedis
from structlog import get_logger

from orchestrator.services.subscriptions import _generate_etag
from orchestrator.settings import app_settings
from orchestrator.utils.json import json_dumps, json_loads

logger = get_logger(__name__)

cache = Redis.from_url(app_settings.CACHE_URI)

ONE_WEEK = 3600 * 24 * 7


def caching_models_enabled() -> bool:
    return getenv("AIOCACHE_DISABLE", "0") == "0" and app_settings.CACHE_DOMAIN_MODELS


def to_redis(subscription: Dict[str, Any]) -> None:
    if caching_models_enabled():
        logger.info("Setting cache for subscription.", subscription=subscription["subscription_id"])
        etag = _generate_etag(subscription)
        cache.set(f"domain:{subscription['subscription_id']}", json_dumps(subscription), ex=ONE_WEEK)
        cache.set(f"domain:etag:{subscription['subscription_id']}", etag, ex=ONE_WEEK)
    else:
        logger.warning("Caching disabled, not caching subscription", subscription=subscription["subscription_id"])


def from_redis(subscription_id: UUID) -> Optional[Tuple[Any, str]]:
    if caching_models_enabled():
        logger.info("Retrieving subscription from cache", subscription=subscription_id)
        obj = cache.get(f"domain:{subscription_id}")
        etag = cache.get(f"domain:etag:{subscription_id}")
        if obj and etag:
            return json_loads(obj), etag.decode("utf-8")
        else:
            return None
    else:
        logger.warning("Caching disabled, not loading subscription", subscription=subscription_id)
        return None


def delete_from_redis(subscription_id: UUID) -> None:
    if caching_models_enabled():
        logger.info("Deleting subscription object from cache", subscription_id=subscription_id)
        cache.delete(f"domain:{subscription_id}")
        cache.delete(f"domain:etag:{subscription_id}")
    else:
        logger.warning("Caching disabled, not deleting subscription", subscription=subscription_id)


async def delete_keys_matching_pattern(_cache: AIORedis, pattern: str, chunksize: int = 5000) -> int:
    """Delete all keys matching the given pattern.

    Usage:
        >>> await delete_keys_matching_pattern("orchestrator:foo:*")  # doctest:+SKIP
    """
    deleted = 0

    async def fetch(_cursor: int = 0) -> tuple[int, list[bytes]]:
        return await _cache.scan(cursor=_cursor, match=pattern, count=chunksize)

    cursor, keys = await fetch()
    while keys or cursor != 0:
        if keys:
            deleted += await _cache.delete(*keys)
        cursor, keys = await fetch(_cursor=cursor)

    logger.debug("Deleted keys matching pattern", pattern=pattern, deleted=deleted)
    return deleted



from typing import Optional
from uuid import uuid4

import pytest

from orchestrator.db import ProductTable, db
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import SubscriptionModel
from orchestrator.types import SubscriptionLifecycle


@pytest.fixture
def test_product_type_sub_one(test_product_sub_block_one):
    SubBlockOneForTestInactive, SubBlockOneForTestProvisioning, SubBlockOneForTest = test_product_sub_block_one

    class ProductSubOneInactive(SubscriptionModel, is_base=True):
        test_block: Optional[SubBlockOneForTestInactive]

    class ProductSubOneProvisioning(ProductSubOneInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
        test_block: SubBlockOneForTestProvisioning

    class ProductSubOne(ProductSubOneProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_block: SubBlockOneForTest

    SUBSCRIPTION_MODEL_REGISTRY["ProductSubOne"] = ProductSubOne
    yield ProductSubOneInactive, ProductSubOneProvisioning, ProductSubOne
    del SUBSCRIPTION_MODEL_REGISTRY["ProductSubOne"]


@pytest.fixture
def test_product_sub_one(test_product_sub_block_one_db):
    product = ProductTable(
        name="ProductSubOne", description="Test ProductSubOne", product_type="Test", tag="Sub", status="active"
    )

    product.product_blocks = [test_product_sub_block_one_db]
    db.session.add(product)
    db.session.commit()
    return product.product_id


@pytest.fixture
def sub_one_subscription_1(test_product_type_sub_one, test_product_sub_block_one, test_product_sub_one):
    ProductSubOneInactive, _, ProductSubOne = test_product_type_sub_one
    SubBlockOneForTestInactive, _, _ = test_product_sub_block_one

    sub_one_subscription_inactive = ProductSubOneInactive.from_product_id(
        product_id=test_product_sub_one, customer_id=uuid4()
    )
    sub_one_subscription_inactive.test_block = SubBlockOneForTestInactive.new(
        subscription_id=sub_one_subscription_inactive.subscription_id, int_field=1, str_field="blah"
    )
    sub_one_subscription_inactive.save()
    sub_one_subscription_active = ProductSubOne.from_other_lifecycle(
        sub_one_subscription_inactive, SubscriptionLifecycle.ACTIVE
    )
    sub_one_subscription_active.save()
    return sub_one_subscription_active



from typing import Callable

import structlog

from orchestrator.db import ProductBlockTable, ProductTable
from orchestrator.db.database import SearchQuery
from orchestrator.db.filters import generic_filter
from orchestrator.db.filters.generic_filters import generic_is_like_filter, generic_values_in_column_filter

logger = structlog.get_logger(__name__)


def product_block_filter(query: SearchQuery, value: str) -> SearchQuery:
    """Filter ProductBlocks by '-'-separated list of Product block 'name' (column) values."""
    blocks = value.split("-")
    return query.filter(ProductTable.product_blocks.any(ProductBlockTable.name.in_(blocks)))


PRODUCT_FILTER_FUNCTIONS_BY_COLUMN: dict[str, Callable[[SearchQuery, str], SearchQuery]] = {
    "product_id": generic_is_like_filter(ProductTable.product_id),
    "name": generic_is_like_filter(ProductTable.name),
    "description": generic_is_like_filter(ProductTable.description),
    "product_type": generic_is_like_filter(ProductTable.product_type),
    "status": generic_values_in_column_filter(ProductTable.status),
    "tag": generic_values_in_column_filter(ProductTable.tag),
    "product_blocks": product_block_filter,
}

filter_products = generic_filter(PRODUCT_FILTER_FUNCTIONS_BY_COLUMN)



from http import HTTPStatus
from uuid import uuid4

import pytest

from orchestrator.db import FixedInputTable, ProductTable, SubscriptionTable, db

PRODUCT_ID = str(uuid4())
SUBSCRIPTION_ID = str(uuid4())
PORT_SPEED_VALUE = "100000"
SERVICE_SPEED_VALUE = "100000"
PORT_SPEED = "port_speed"
SERVICE_SPEED = "service_speed"


@pytest.fixture
def seed():
    FixedInputTable.query.delete()
    fixed_inputs = [
        FixedInputTable(name=PORT_SPEED, value=PORT_SPEED_VALUE),
        FixedInputTable(name=SERVICE_SPEED, value=SERVICE_SPEED_VALUE),
    ]
    product = ProductTable(
        product_id=PRODUCT_ID,
        name="ProductTable",
        description="description",
        product_type="Port",
        tag="SSP",
        status="active",
        fixed_inputs=fixed_inputs,
    )
    subscription = SubscriptionTable(
        subscription_id=SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=product,
        customer_id=uuid4(),
    )

    db.session.add(product)
    db.session.add(subscription)
    db.session.commit()


def test_configuration(seed, test_client):
    response = test_client.get("/api/fixed_inputs/configuration")
    assert HTTPStatus.OK == response.status_code
    config = response.json()
    fixed_inputs_config = list(map(lambda x: x["name"], config["fixed_inputs"]))
    for tag, fixed_inputs in config["by_tag"].items():
        for fi in fixed_inputs:
            name = list(fi.keys())[0]
            assert name in fixed_inputs_config, f"{tag} has non-existent fixed_input: {name}"



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from http import HTTPStatus
from typing import Any, Callable, Dict, List, Optional
from uuid import UUID

from orchestrator import app_settings
from orchestrator.api.error_handling import raise_status
from orchestrator.db import ProcessTable
from orchestrator.services.processes import create_process
from orchestrator.targets import Target
from orchestrator.types import State
from orchestrator.workflows import get_workflow

SYSTEM_USER = "SYSTEM"


def _celery_start_process(
    workflow_key: str, user_inputs: Optional[List[State]], user: str = SYSTEM_USER, **kwargs: Any
) -> UUID:
    """Client side call of Celery."""
    from orchestrator.services.tasks import NEW_TASK, NEW_WORKFLOW, get_celery_task

    workflow = get_workflow(workflow_key)
    if not workflow:
        raise_status(HTTPStatus.NOT_FOUND, "Workflow does not exist")

    task_name = NEW_TASK if workflow.target == Target.SYSTEM else NEW_WORKFLOW
    trigger_task = get_celery_task(task_name)
    pstat = create_process(workflow_key, user_inputs, user)

    tasks = pstat.state.s
    result = trigger_task.delay(pstat.pid, workflow_key, tasks, user)

    # Enables "Sync celery tasks. This will let the app wait until celery completes"
    if app_settings.TESTING:
        pid = result.get()
        if not pid:
            raise RuntimeError("Celery worker has failed to resume process")

    return pstat.pid


def _celery_resume_process(
    process: ProcessTable,
    *,
    user_inputs: Optional[List[State]],
    user: Optional[str],
    **kwargs: Any,
) -> UUID:
    """Client side call of Celery."""
    from orchestrator.services.processes import load_process
    from orchestrator.services.tasks import RESUME_TASK, RESUME_WORKFLOW, get_celery_task

    pstat = load_process(process)
    workflow = pstat.workflow

    task_name = RESUME_TASK if workflow.target == Target.SYSTEM else RESUME_WORKFLOW
    trigger_task = get_celery_task(task_name)
    result = trigger_task.delay(pstat.pid, user_inputs, user)

    _celery_set_process_status_resumed(process)

    # Enables "Sync celery tasks. This will let the app wait until celery completes"
    if app_settings.TESTING:
        pid = result.get()
        if not pid:
            raise RuntimeError("Celery worker has failed to resume process")

    return pstat.pid


def _celery_set_process_status_resumed(process: ProcessTable) -> None:
    """Set the process status to RESUMED to prevent re-adding to task queue.

    Args:
        process: Process from database
    """
    from orchestrator.db import db
    from orchestrator.workflow import ProcessStatus

    process.last_status = ProcessStatus.RESUMED
    db.session.add(process)
    db.session.commit()


def _celery_validate(validation_workflow: str, json: Optional[List[State]]) -> None:
    _celery_start_process(validation_workflow, user_inputs=json)


CELERY_EXECUTION_CONTEXT: Dict[str, Callable] = {
    "start": _celery_start_process,
    "resume": _celery_resume_process,
    "validate": _celery_validate,
}



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Union

from broadcaster import Broadcast
from fastapi import WebSocket, status
from starlette.concurrency import run_until_first_complete
from structlog import get_logger

from orchestrator.utils.json import json_dumps, json_loads

logger = get_logger(__name__)


class BroadcastWebsocketManager:
    def __init__(self, broadcast_url: str):
        self.connected: list[WebSocket] = []
        self.sub_broadcast = Broadcast(broadcast_url)
        self.pub_broadcast = Broadcast(broadcast_url)

    async def connect_redis(self) -> None:
        await self.sub_broadcast.connect()

    async def disconnect_redis(self) -> None:
        await self.sub_broadcast.disconnect()

    async def connect(self, websocket: WebSocket, channel: str) -> None:
        self.connected.append(websocket)
        self.log_amount_of_connections()
        try:
            await run_until_first_complete(
                (self.sender, {"websocket": websocket, "channel": channel}),
                (self.receiver, {"websocket": websocket, "channel": channel}),
            )
        except Exception:  # noqa: S110
            pass
        self.remove_ws_from_connected_list(websocket)

    async def disconnect(
        self, websocket: WebSocket, code: int = status.WS_1000_NORMAL_CLOSURE, reason: Union[Dict, str, None] = None
    ) -> None:
        if reason:
            await websocket.send_text(json_dumps(reason))
        await websocket.close(code)
        self.remove_ws_from_connected_list(websocket)

    async def disconnect_all(self) -> None:
        for websocket in self.connected:
            await self.disconnect(websocket)

    async def receiver(self, websocket: WebSocket, channel: str) -> None:
        async for message in websocket.iter_text():
            if message == "__ping__":
                await websocket.send_text("__pong__")

    async def sender(self, websocket: WebSocket, channel: str) -> None:
        async with self.sub_broadcast.subscribe(channel=channel) as subscriber:
            async for event in subscriber:
                await websocket.send_text(event.message)

                json = json_loads(event.message)
                if type(json) is dict and "close" in json and json["close"] and channel != "processes":
                    await self.disconnect(websocket)
                    break

    async def broadcast_data(self, channels: List[str], data: Dict) -> None:
        await self.pub_broadcast.connect()
        for channel in channels:
            await self.pub_broadcast.publish(channel, message=json_dumps(data))
        await self.pub_broadcast.disconnect()

    def remove_ws_from_connected_list(self, websocket: WebSocket) -> None:
        if websocket in self.connected:
            self.connected.remove(websocket)
        self.log_amount_of_connections()

    def log_amount_of_connections(self) -> None:
        logger.info("Websocket Connections: %s", len(self.connected))



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import structlog
from sqlalchemy.orm import defer, joinedload

from orchestrator.db import ProcessSubscriptionTable, ProcessTable, SubscriptionTable
from orchestrator.db.filters import Filter
from orchestrator.db.filters.process import filter_processes
from orchestrator.db.range import apply_range_to_query
from orchestrator.db.sorting import Sort
from orchestrator.db.sorting.process import sort_processes
from orchestrator.graphql.pagination import Connection, PageInfo
from orchestrator.graphql.schemas.process import ProcessGraphqlSchema, ProcessType
from orchestrator.graphql.types import CustomInfo, GraphqlFilter, GraphqlSort
from orchestrator.graphql.utils.create_resolver_error_handler import create_resolver_error_handler
from orchestrator.services.processes import load_process
from orchestrator.utils.show_process import show_process

logger = structlog.get_logger(__name__)


def enrich_process(process: ProcessTable) -> ProcessGraphqlSchema:
    p = load_process(process)
    data = show_process(process, p)
    return ProcessGraphqlSchema(**data)


async def resolve_processes(
    info: CustomInfo,
    filter_by: Union[list[GraphqlFilter], None] = None,
    sort_by: Union[list[GraphqlSort], None] = None,
    first: int = 10,
    after: int = 0,
) -> Connection[ProcessType]:
    _error_handler = create_resolver_error_handler(info)
    pydantic_filter_by: list[Filter] = [item.to_pydantic() for item in filter_by] if filter_by else []
    pydantic_sort_by: list[Sort] = [item.to_pydantic() for item in sort_by] if sort_by else []
    logger.info("resolve_processes() called", range=[after, after + first], sort=sort_by, filter=pydantic_filter_by)

    # the joinedload on ProcessSubscriptionTable.subscription via ProcessBaseSchema.process_subscriptions prevents a query for every subscription later.
    # tracebacks are not presented in the list of processes and can be really large.
    query = ProcessTable.query.options(
        joinedload(ProcessTable.process_subscriptions)
        .joinedload(ProcessSubscriptionTable.subscription)
        .joinedload(SubscriptionTable.product),
        defer("traceback"),
    )

    query = filter_processes(query, pydantic_filter_by, _error_handler)
    query = sort_processes(query, pydantic_sort_by, _error_handler)
    total = query.count()
    query = apply_range_to_query(query, after, first)

    processes = query.all()
    has_next_page = len(processes) > first

    # exclude last item as it was fetched to know if there is a next page
    processes = processes[:first]
    processes_length = len(processes)
    start_cursor = after if processes_length else None
    end_cursor = after + processes_length - 1
    page_processes = [ProcessType.from_pydantic(enrich_process(p)) for p in processes]

    return Connection(
        page=page_processes,
        page_info=PageInfo(
            has_previous_page=bool(after),
            has_next_page=has_next_page,
            start_cursor=start_cursor,
            end_cursor=end_cursor,
            total_items=total if total else None,
        ),
    )



# Copyright 2022 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from http import HTTPStatus
from typing import Union

process_fields = [
    "assignee",
    "createdBy",
    "failedReason",
    "isTask",
    "lastStep",
    "traceback",
    "id",
    "lastModified",
    "started",
    "workflowName",
    "status",
    "step",
    "product",
]


def get_processes_query(
    first: int = 10,
    after: int = 0,
    filter_by: Union[list[str], None] = None,
    sort_by: Union[list[dict[str, str]], None] = None,
) -> bytes:
    query = """
query ProcessQuery($first: Int!, $after: Int!, $sortBy: [GraphqlSort!], $filterBy: [GraphqlFilter!]) {
  processes(first: $first, after: $after, sortBy: $sortBy, filterBy: $filterBy) {
    page {
      assignee
      createdBy
      failedReason
      isTask
      lastStep
      traceback
      id
      lastModified
      started
      workflowName
      status
      step
      product
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
}
    """
    return json.dumps(
        {
            "operationName": "ProcessQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
            },
        }
    ).encode("utf-8")


def get_processes_query_with_subscriptions(
    first: int = 10,
    after: int = 0,
    filter_by: Union[list[str], None] = None,
    sort_by: Union[list[dict[str, str]], None] = None,
) -> bytes:
    query = """
query ProcessQuery($first: Int!, $after: Int!, $sortBy: [GraphqlSort!], $filterBy: [GraphqlFilter!]) {
  processes(first: $first, after: $after, sortBy: $sortBy, filterBy: $filterBy) {
    page {
      assignee
      createdBy
      failedReason
      isTask
      lastStep
      traceback
      id
      lastModified
      started
      workflowName
      status
      step
      product
      subscriptions {
        page {
          description
          subscriptionId
          productId
          status
          insync
          note
          startDate
          endDate
        }
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
}
    """
    return json.dumps(
        {
            "operationName": "ProcessQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
            },
        }
    ).encode("utf-8")


def test_processes_has_next_page(
    test_client, mocked_processes, mocked_processes_resumeall, generic_subscription_2, generic_subscription_1
):
    data = get_processes_query()
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    assert len(processes) == 10

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": "19",
    }

    for process in processes:
        for field in process_fields:
            assert field in process


def test_process_has_previous_page(
    test_client, mocked_processes, mocked_processes_resumeall, generic_subscription_2, generic_subscription_1
):
    data = get_processes_query(after=1)
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    assert len(processes) == 10

    assert pageinfo == {
        "hasPreviousPage": True,
        "hasNextPage": True,
        "startCursor": 1,
        "endCursor": 10,
        "totalItems": "19",
    }


def test_processes_sorting_asc(
    test_client, mocked_processes, mocked_processes_resumeall, generic_subscription_2, generic_subscription_1
):
    # when

    data = get_processes_query(sort_by=[{"field": "started", "order": "ASC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": "19",
    }

    assert processes[0]["started"] == "2020-01-14T09:30:00+00:00"
    assert processes[1]["started"] == "2020-01-14T09:30:00+00:00"
    assert processes[2]["started"] == "2020-01-15T09:30:00+00:00"


def test_processes_sorting_desc(
    test_client, mocked_processes, mocked_processes_resumeall, generic_subscription_2, generic_subscription_1
):
    # when

    data = get_processes_query(sort_by=[{"field": "started", "order": "DESC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": "19",
    }

    assert processes[0]["started"] == "2020-01-19T09:30:00+00:00"
    assert processes[1]["started"] == "2020-01-19T09:30:00+00:00"
    assert processes[2]["started"] == "2020-01-19T09:30:00+00:00"
    assert processes[3]["started"] == "2020-01-19T09:30:00+00:00"
    assert processes[4]["started"] == "2020-01-18T09:30:00+00:00"


def test_processes_has_filtering(
    test_client, mocked_processes, mocked_processes_resumeall, generic_subscription_2, generic_subscription_1
):
    # when

    data = get_processes_query(filter_by=[{"field": "status", "value": "completed"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    # then

    assert HTTPStatus.OK == response.status_code

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 3,
        "totalItems": "4",
    }

    for process in processes:
        assert process["status"] == "COMPLETED"


def test_processes_filtering_with_invalid_filter(
    test_client, mocked_processes, mocked_processes_resumeall, generic_subscription_2, generic_subscription_1
):
    # when

    data = get_processes_query(
        filter_by=[{"field": "status", "value": "completed"}, {"field": "test", "value": "invalid"}]
    )
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    errors = result["errors"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    assert errors == [
        {
            "message": "Invalid filter arguments",
            "path": [None, "processes", "Query"],
            "extensions": {
                "invalid_filters": [{"field": "test", "value": "invalid"}],
                "valid_filter_keys": [
                    "pid",
                    "istask",
                    "assignee",
                    "status",
                    "workflow",
                    "creator",
                    "product",
                    "tag",
                    "subscription",
                    "subscriptionId",
                    "target",
                ],
            },
        }
    ]
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 3,
        "totalItems": "4",
    }

    for process in processes:
        assert process["status"] == "COMPLETED"


def test_single_process(
    test_client, mocked_processes, mocked_processes_resumeall, generic_subscription_2, generic_subscription_1
):
    process_pid = str(mocked_processes[0])
    # when

    data = get_processes_query(filter_by=[{"field": "pid", "value": process_pid}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    assert len(processes) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": "1",
    }
    assert processes[0]["id"] == process_pid


def test_single_process_with_subscriptions(
    test_client, mocked_processes, mocked_processes_resumeall, generic_subscription_2, generic_subscription_1
):
    process_pid = str(mocked_processes[0])
    # when

    data = get_processes_query_with_subscriptions(filter_by=[{"field": "pid", "value": process_pid}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"]["processes"]
    processes = processes_data["page"]
    pageinfo = processes_data["pageInfo"]

    assert len(processes) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": "1",
    }
    assert processes[0]["id"] == process_pid
    assert processes[0]["subscriptions"]["page"][0]["subscriptionId"] == generic_subscription_1



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from orchestrator.db import db
from orchestrator.forms import FormPage
from orchestrator.forms.validators import LongText
from orchestrator.services import subscriptions
from orchestrator.settings import app_settings
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, UUIDstr
from orchestrator.utils.json import to_serializable
from orchestrator.workflow import StepList, conditional, done, init, workflow
from orchestrator.workflows.steps import cache_domain_models, store_process_subscription
from orchestrator.workflows.utils import wrap_modify_initial_input_form


def initial_input_form(subscription_id: UUIDstr) -> FormGenerator:
    subscription = subscriptions.get_subscription(subscription_id)
    subscription_backup = {subscription_id: to_serializable(subscription)}
    old_note = subscription.note

    class ModifyNoteForm(FormPage):
        note: LongText = old_note

    user_input = yield ModifyNoteForm
    subscription.note = user_input.note
    db.session.add(subscription)
    return {
        "old_note": old_note,
        "note": user_input.note,
        "subscription": to_serializable(subscription),
        "__old_subscriptions__": subscription_backup,
    }


@workflow("Modify Note", initial_input_form=wrap_modify_initial_input_form(initial_input_form), target=Target.MODIFY)
def modify_note() -> StepList:
    push_subscriptions = conditional(lambda _: app_settings.CACHE_DOMAIN_MODELS)
    return init >> store_process_subscription(Target.MODIFY) >> push_subscriptions(cache_domain_models) >> done



# Copyright 2019-2022 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional, Union

from pydantic import RedisDsn

from orchestrator.distlock.managers.memory_distlock_manager import Lock as MemoryLock
from orchestrator.distlock.managers.memory_distlock_manager import MemoryDistLockManager
from orchestrator.distlock.managers.redis_distlock_manager import Lock as RedisLock
from orchestrator.distlock.managers.redis_distlock_manager import RedisDistLockManager

DistLock = Union[MemoryLock, RedisLock]


class DistLockManager:
    """Provides an interface to lock access to a resource in a distributed system.

    The lock is advisory; it is up to the caller to respect it.

    Creating a lock is non-blocking, it succeeds or fails immediately.

    Locks are to be created with an expiration period after which the backend
    implementation must release it.
    """

    _backend: Union[MemoryDistLockManager, RedisDistLockManager]

    def __init__(self, enabled: bool, backend: Optional[str] = None, redis_dsn: Optional[RedisDsn] = None):
        self.enabled = enabled
        self.connected = False
        if backend == "redis" and redis_dsn:
            self._backend = RedisDistLockManager(redis_dsn)
        else:
            self._backend = MemoryDistLockManager()

    async def connect_redis(self) -> None:
        if not self.connected:
            await self._backend.connect_redis()
            self.connected = True

    async def disconnect_redis(self) -> None:
        if self.connected:
            await self._backend.disconnect_redis()
            self.connected = False

    async def get_lock(self, resource: str, expiration_seconds: int) -> Optional[DistLock]:
        return await self._backend.get_lock(resource, expiration_seconds)

    async def release_lock(self, resource: DistLock) -> None:
        await self._backend.release_lock(resource)  # type: ignore

    def release_sync(self, resource: DistLock) -> None:
        self._backend.release_sync(resource)  # type: ignore



# Copyright 2022-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Map some Orchestrator types to scalars
from typing import Callable

import strawberry
from graphql import GraphQLError
from oauth2_lib.fastapi import OIDCUserModel
from strawberry.fastapi import BaseContext
from strawberry.types import Info
from strawberry.types.info import RootValueType

from orchestrator.db.filters import Filter
from orchestrator.db.sorting import Sort, SortOrder


class CustomContext(BaseContext):
    def __init__(self, get_current_user: Callable[[], OIDCUserModel], get_opa_decision: Callable[[str], bool]):
        self.errors: list[GraphQLError] = []
        self.get_current_user = get_current_user
        self.get_opa_decision = get_opa_decision
        super().__init__()


CustomInfo = Info[CustomContext, RootValueType]


@strawberry.experimental.pydantic.input(model=Sort)
class GraphqlSort:
    field: str = strawberry.field(description="Field to sort on")
    order: SortOrder = strawberry.field(default=SortOrder.ASC, description="Sort order (ASC or DESC")


@strawberry.experimental.pydantic.input(model=Filter)
class GraphqlFilter:
    field: str = strawberry.field(description="Field to filter on")
    value: str = strawberry.field(description="Value to sort the field on")



import pytest

from orchestrator.db import FixedInputTable, ProductTable, db
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import ProductModel, SubscriptionModel
from orchestrator.domain.lifecycle import ProductLifecycle
from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.fixtures.products.product_blocks.product_block_one_nested import (
    ProductBlockOneNestedForTest,
    ProductBlockOneNestedForTestInactive,
    ProductBlockOneNestedForTestProvisioning,
)


@pytest.fixture
def test_product_type_one_nested():
    class ProductTypeOneNestedForTestInactive(SubscriptionModel, is_base=True):
        test_fixed_input: bool
        block: ProductBlockOneNestedForTestInactive

    class ProductTypeOneNestedForTestProvisioning(
        ProductTypeOneNestedForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
    ):
        test_fixed_input: bool
        block: ProductBlockOneNestedForTestProvisioning

    class ProductTypeOneNestedForTest(
        ProductTypeOneNestedForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        test_fixed_input: bool
        block: ProductBlockOneNestedForTest

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOneNested"] = ProductTypeOneNestedForTest
    yield ProductTypeOneNestedForTestInactive, ProductTypeOneNestedForTestProvisioning, ProductTypeOneNestedForTest
    del SUBSCRIPTION_MODEL_REGISTRY["TestProductOneNested"]


@pytest.fixture
def test_product_one_nested(test_product_block_one_nested_db_in_use_by_block):
    product = ProductTable(
        name="TestProductOneNested", description="Test ProductTable", product_type="Test", tag="TEST", status="active"
    )

    fixed_input = FixedInputTable(name="test_fixed_input", value="False")

    product_block = test_product_block_one_nested_db_in_use_by_block
    product.fixed_inputs = [fixed_input]
    product.product_blocks = [product_block]

    db.session.add(product)
    db.session.commit()

    return product.product_id


@pytest.fixture
def test_product_model_nested(test_product_one_nested):
    return ProductModel(
        product_id=test_product_one_nested,
        name="TestProductOneNested",
        description="Test ProductTable",
        product_type="Test",
        tag="TEST",
        status=ProductLifecycle.ACTIVE,
    )



from datetime import datetime
from typing import TYPE_CHECKING, Annotated, Any, Dict, List, Optional, Union
from uuid import UUID

import strawberry
from oauth2_lib.graphql_authentication import authenticated_field
from sqlalchemy.orm import load_only
from strawberry.scalars import JSON

from orchestrator.config.assignee import Assignee
from orchestrator.db import ProcessTable
from orchestrator.graphql.pagination import EMPTY_PAGE, Connection
from orchestrator.graphql.types import CustomInfo, GraphqlFilter, GraphqlSort
from orchestrator.schemas.base import OrchestratorBaseModel
from orchestrator.schemas.process import ProcessForm, ProcessStepSchema
from orchestrator.workflow import ProcessStatus

if TYPE_CHECKING:
    from orchestrator.graphql.schemas.subscription import SubscriptionType


# TODO: Change to the orchestrator.schemas.process version when subscriptions are typed in strawberry.
class ProcessBaseSchema(OrchestratorBaseModel):
    id: UUID
    workflow_name: str
    product: Optional[UUID]
    customer: Optional[UUID]
    assignee: Assignee
    failed_reason: Optional[str]
    traceback: Optional[str]
    step: Optional[str]
    status: ProcessStatus
    last_step: Optional[str]
    created_by: Optional[str]
    started: datetime
    last_modified: datetime
    is_task: bool


class ProcessGraphqlSchema(ProcessBaseSchema):
    current_state: Dict[str, Any]
    steps: List[ProcessStepSchema]
    form: Optional[ProcessForm]


@strawberry.experimental.pydantic.type(model=ProcessForm)
class ProcessFormType:
    title: strawberry.auto
    type: strawberry.auto
    additionalProperties: strawberry.auto
    required: strawberry.auto
    properties: JSON
    definitions: Union[JSON, None]


@strawberry.experimental.pydantic.type(model=ProcessStepSchema)
class ProcessStepType:
    stepid: strawberry.auto
    name: strawberry.auto
    status: strawberry.auto
    created_by: strawberry.auto
    executed: strawberry.auto
    commit_hash: strawberry.auto
    state: Union[JSON, None]


@strawberry.experimental.pydantic.type(model=ProcessGraphqlSchema)
class ProcessType:
    id: strawberry.auto
    workflow_name: strawberry.auto
    product: strawberry.auto
    customer: strawberry.auto
    assignee: strawberry.auto
    failed_reason: strawberry.auto
    traceback: strawberry.auto
    step: strawberry.auto
    status: strawberry.auto
    last_step: strawberry.auto
    created_by: strawberry.auto
    started: strawberry.auto
    last_modified: strawberry.auto
    is_task: strawberry.auto
    steps: strawberry.auto
    form: strawberry.auto
    current_state: Union[JSON, None]

    @authenticated_field(description="Returns list of subscriptions of the process")  # type: ignore
    async def subscriptions(
        self,
        info: CustomInfo,
        filter_by: Union[list[GraphqlFilter], None] = None,
        sort_by: Union[list[GraphqlSort], None] = None,
        first: int = 10,
        after: int = 0,
    ) -> Connection[Annotated["SubscriptionType", strawberry.lazy(".subscription")]]:
        from orchestrator.graphql.resolvers.subscription import resolve_subscriptions

        process: ProcessTable = ProcessTable.query.options(load_only(ProcessTable.process_subscriptions)).get(self.id)
        subscription_ids = [str(s.subscription_id) for s in process.process_subscriptions]
        if not subscription_ids:
            return EMPTY_PAGE
        filter_by_with_related_subscriptions = (filter_by or []) + [
            GraphqlFilter(field="subscriptionIds", value=",".join(subscription_ids))
        ]
        return await resolve_subscriptions(info, filter_by_with_related_subscriptions, sort_by, first, after)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import queue
import threading
from concurrent.futures.thread import ThreadPoolExecutor
from functools import partial
from http import HTTPStatus
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple
from uuid import UUID, uuid4

import structlog
from deepmerge import Merger
from fastapi import Request
from nwastdlib.ex import show_ex
from sqlalchemy.orm import joinedload

from orchestrator.api.error_handling import raise_status
from orchestrator.config.assignee import Assignee
from orchestrator.db import EngineSettingsTable, ProcessStepTable, ProcessSubscriptionTable, ProcessTable, db
from orchestrator.distlock import distlock_manager
from orchestrator.forms import FormValidationError, post_process
from orchestrator.schemas.engine_settings import WorkerStatus
from orchestrator.settings import ExecutorType, app_settings
from orchestrator.targets import Target
from orchestrator.types import BroadcastFunc, State
from orchestrator.utils.datetime import nowtz
from orchestrator.utils.errors import error_state_to_dict
from orchestrator.websocket import (
    WS_CHANNELS,
    create_process_websocket_data,
    send_process_data_to_websocket,
    websocket_manager,
)
from orchestrator.websocket.websocket_manager import WebSocketManager
from orchestrator.workflow import Failed
from orchestrator.workflow import Process as WFProcess
from orchestrator.workflow import ProcessStat, ProcessStatus, Step, StepList, Success, Workflow, abort_wf, runwf
from orchestrator.workflows import get_workflow
from orchestrator.workflows.removed_workflow import removed_workflow

logger = structlog.get_logger(__name__)

StateMerger = Merger([(dict, ["merge"])], ["override"], ["override"])

SYSTEM_USER = "SYSTEM"

_workflow_executor = None


def get_execution_context() -> Dict[str, Callable]:
    from orchestrator.services.celery import CELERY_EXECUTION_CONTEXT

    return CELERY_EXECUTION_CONTEXT if app_settings.EXECUTOR == ExecutorType.WORKER else THREADPOOL_EXECUTION_CONTEXT


def get_thread_pool() -> ThreadPoolExecutor:
    """
    Get and optionally initialise a ThreadPoolExecutor.

    Returns:
        ThreadPoolExecutor

    """
    global _workflow_executor
    if _workflow_executor is None:
        _workflow_executor = ThreadPoolExecutor(max_workers=app_settings.MAX_WORKERS)
    return _workflow_executor


def shutdown_thread_pool() -> None:
    """Gracefully shutdown existing ThreadPoolExecutor and delete it."""
    global _workflow_executor
    if isinstance(_workflow_executor, ThreadPoolExecutor):
        _workflow_executor.shutdown(wait=True)
        _workflow_executor = None


def _db_create_process(stat: ProcessStat) -> None:
    p = ProcessTable(
        pid=stat.pid,
        workflow=stat.workflow.name,
        last_status=ProcessStatus.CREATED,
        created_by=stat.current_user,
        is_task=stat.workflow.target == Target.SYSTEM,
    )
    db.session.add(p)
    db.session.commit()


def _db_log_step(
    stat: ProcessStat,
    step: Step,
    process_state: WFProcess,
    broadcast_func: Optional[BroadcastFunc] = None,
) -> WFProcess:
    """Write the current step to the db.

    Args:
        stat: ProcessStat of process
        step: Current step
        process_state: State of process after current step
        broadcast_func: Optional function to broadcast process data

    Returns:
        WFProcess

    """
    p = ProcessTable.query.get(stat.pid)
    if p is None:
        raise ValueError(f"Failed to write failure step to process: process with PID {stat.pid} not found")

    p.last_step = step.name
    p.last_status = process_state.overall_status
    p.assignee = step.assignee

    step_state: State = process_state.unwrap()
    current_step = None
    if process_state.isfailed() or process_state.iswaiting():
        failed_reason = step_state.get("error")
        failed_details = step_state.get("details")
        # pop also removes the traceback from the dict
        traceback = step_state.pop("traceback", None)

        p.failed_reason = failed_reason
        p.traceback = traceback

        if process_state.isfailed() and p.is_task:
            # Check if we need a special failed status:
            # If it is an AssertionError:
            if step_state.get("class") == "AssertionError" or step_state.get("class") == "InconsistentData":
                p.assignee = Assignee.NOC
                p.last_status = ProcessStatus.INCONSISTENT_DATA
            # If we encounter a connectivity issue with an underlying api:
            elif step_state.get("class") == "MaxRetryError" or (
                step_state.get("class") == "ApiException"
                and step_state.get("status_code")
                in (
                    HTTPStatus.BAD_GATEWAY,
                    HTTPStatus.SERVICE_UNAVAILABLE,
                    HTTPStatus.GATEWAY_TIMEOUT,
                )
            ):
                p.assignee = Assignee.SYSTEM
                p.last_status = ProcessStatus.API_UNAVAILABLE
            else:
                p.assignee = Assignee.SYSTEM

        # check if last error state is identical to determine if we add a new step or update the last one
        last_db_step = p.steps[-1] if len(p.steps) else None

        if (
            last_db_step is not None
            and last_db_step.status == process_state.status
            and last_db_step.name == step.name
            and failed_reason == last_db_step.state.get("error")
            and failed_details == last_db_step.state.get("details")
        ):
            current_step = last_db_step

    else:
        p.failed_reason = None
        p.traceback = None

    db.session.add(p)

    if current_step is None:
        # add a new entry to the process stat
        logger.info("Adding a new process step with state info")
        current_step = ProcessStepTable(
            pid=stat.pid,
            name=step.name,
            status=process_state.status,
            state=step_state,
            created_by=stat.current_user,
        )
    else:
        # update the last one with the repeated info
        retries = current_step.state.get("retries", 0) + 1
        executed_at = current_step.state.get("executed_at", [])
        executed_at.append(str(current_step.executed_at))

        # write new state info and execution date
        current_step.state = step_state | {
            "retries": retries,
            "executed_at": executed_at,
        }
        logger.info(
            "Updating existing process step with state info about the error",
            retries=retries,
        )

    # Always explicitly set this instead of leaving it to the database to prevent failing tests
    # Test will fail if multiple steps have the same timestamp
    current_step.executed_at = nowtz()

    db.session.add(current_step)
    try:
        db.session.commit()
    except BaseException:
        db.session.rollback()
        raise

    if websocket_manager.enabled:
        new_pStat = load_process(p)
        websocket_data = create_process_websocket_data(p, new_pStat)
        send_process_data_to_websocket(p.pid, websocket_data, broadcast_func=broadcast_func)

    # Return the state as stored in the database
    return process_state.__class__(current_step.state)


def safe_logstep(
    stat: ProcessStat,
    step: Step,
    process_state: WFProcess,
    broadcast_func: Optional[BroadcastFunc] = None,
) -> WFProcess:
    """Log step and handle failures in logging.

    We need to be robust in failures to write steps to database. If that happens we try again with the failure.
    If that is also failing we give up by raising an exception which should be caught and written by _db_log_process_ex
    """

    try:
        return _db_log_step(stat, step, process_state, broadcast_func=broadcast_func)
    except Exception as e:
        logger.exception("Failed to save step", stat=stat, step=step, process_state=process_state)
        failure = Failed(error_state_to_dict(e))

        # Try writing the failure, but return the original exception on success
        # on a second failure the exception should be handled higher
        return _db_log_step(stat, step, failure, broadcast_func=broadcast_func)


def _db_log_process_ex(pid: UUID, ex: Exception) -> None:
    """
    Write the exception to the process or task when everything else has failed.

    Args:
        pid: the pid of the workflow process
        ex: the Exception message

    Returns: None, there is no one to listen at this point

    """

    p = ProcessTable.query.get(pid)
    if p is None:
        logger.error(
            "Failed to write failure to database: Process with PID %s not found",
            pid,
            pid=pid,
        )
        return

    logger.warning("Writing only process state to DB as step couldn't be found", pid=pid)
    p.last_step = "Unknown"
    if p.last_status != ProcessStatus.WAITING:
        p.last_status = ProcessStatus.FAILED
    p.failed_reason = str(ex)
    p.traceback = show_ex(ex)
    db.session.add(p)
    try:
        db.session.commit()
    except BaseException:
        logger.exception("Commit failed, rolling back", pid=pid)
        db.session.rollback()
        raise


def _get_process(pid: UUID) -> ProcessTable:
    process = ProcessTable.query.options(
        joinedload(ProcessTable.steps),
        joinedload(ProcessTable.process_subscriptions).joinedload(ProcessSubscriptionTable.subscription),
    ).get(pid)

    if not process:
        raise_status(HTTPStatus.NOT_FOUND, f"Process with pid {pid} not found")

    return process


def _run_process_async(pid: UUID, f: Callable) -> UUID:
    def _update_running_processes(method: Literal["+", "-"], *args: Any) -> None:
        """
        Update amount of running processes by one.

        Args:
            method: Add or subtract by one the amount of running processes
            args: Any args that are still going to be passed. When called as a callback this will be the future.
        Returns:
            None

        """
        engine_settings = EngineSettingsTable.query.with_for_update().one()
        engine_settings.running_processes += 1 if method == "+" else -1
        if engine_settings.running_processes < 0:
            engine_settings.running_processes = 0
        db.session.commit()

    def run() -> WFProcess:
        try:
            with db.database_scope():
                try:
                    logger.new(process_id=str(pid))
                    result = f()
                except Exception as ex:
                    # We still have access to the database, so we can log at least something
                    _db_log_process_ex(pid, ex)
                    raise
                finally:
                    _update_running_processes("-")
        except Exception as ex:
            # We lost access to database here, so we can only log
            logger.exception("Unknown workflow failure", pid=pid)
            result = Failed(ex)

        return result

    _update_running_processes("+")
    if app_settings.EXECUTOR == ExecutorType.THREADPOOL:
        workflow_executor = get_thread_pool()
        process_handle = workflow_executor.submit(run)

        # Wait for the thread to return.
        if app_settings.TESTING:
            process_handle.result()
    elif app_settings.EXECUTOR == ExecutorType.WORKER:
        # No need to run in a thread, just run.
        run()
    else:
        raise RuntimeError("Unknown Executor type")
    return pid


def create_process(
    workflow_key: str,
    user_inputs: Optional[List[State]] = None,
    user: str = SYSTEM_USER,
) -> ProcessStat:
    # ATTENTION!! When modifying this function make sure you make similar changes to `run_workflow` in the test code

    if user_inputs is None:
        user_inputs = [{}]

    pid = uuid4()
    workflow = get_workflow(workflow_key)

    if not workflow:
        raise_status(HTTPStatus.NOT_FOUND, "Workflow does not exist")

    initial_state = {
        "process_id": pid,
        "reporter": user,
        "workflow_name": workflow_key,
        "workflow_target": workflow.target,
    }

    try:
        state = post_process(workflow.initial_input_form, initial_state, user_inputs)
    except FormValidationError:
        logger.exception("Validation errors", user_inputs=user_inputs)
        raise

    pstat = ProcessStat(
        pid,
        workflow=workflow,
        state=Success(state | initial_state),
        log=workflow.steps,
        current_user=user,
    )

    _db_create_process(pstat)

    return pstat


def thread_start_process(
    workflow_key: str,
    user_inputs: Optional[List[State]] = None,
    user: str = SYSTEM_USER,
    broadcast_func: Optional[BroadcastFunc] = None,
) -> UUID:
    pstat = create_process(workflow_key, user_inputs=user_inputs, user=user)

    _safe_logstep_with_func = partial(safe_logstep, broadcast_func=broadcast_func)
    return _run_process_async(pstat.pid, lambda: runwf(pstat, _safe_logstep_with_func))


def start_process(
    workflow_key: str,
    user_inputs: Optional[List[State]] = None,
    user: str = SYSTEM_USER,
    broadcast_func: Optional[BroadcastFunc] = None,
) -> UUID:
    """Start a process for workflow.

    Args:
        workflow_key: name of workflow
        user_inputs: List of form inputs from frontend
        user: User who starts this process
        broadcast_func: Optional function to broadcast process data

    Returns:
        process id

    """
    start_func = get_execution_context()["start"]
    return start_func(workflow_key, user_inputs=user_inputs, user=user, broadcast_func=broadcast_func)


def thread_resume_process(
    process: ProcessTable,
    *,
    user_inputs: Optional[List[State]] = None,
    user: Optional[str] = None,
    broadcast_func: Optional[BroadcastFunc] = None,
) -> UUID:
    # ATTENTION!! When modifying this function make sure you make similar changes to `resume_workflow` in the test code

    if user_inputs is None:
        user_inputs = [{}]

    pstat = load_process(process)

    if pstat.workflow == removed_workflow:
        raise ValueError("This workflow cannot be resumed")

    form = pstat.log[0].form

    user_input = post_process(form, pstat.state.unwrap(), user_inputs)

    if user:
        pstat.update(current_user=user)

    if user_input:
        pstat.update(state=pstat.state.map(lambda state: StateMerger.merge(state, user_input)))

    # enforce an update to the process status to properly show the process
    process.last_status = ProcessStatus.RUNNING
    db.session.add(process)
    db.session.commit()

    _safe_logstep_prep = partial(safe_logstep, broadcast_func=broadcast_func)
    return _run_process_async(pstat.pid, lambda: runwf(pstat, _safe_logstep_prep))


def thread_validate_workflow(validation_workflow: str, json: Optional[List[State]]) -> UUID:
    return thread_start_process(validation_workflow, user_inputs=json)


THREADPOOL_EXECUTION_CONTEXT: Dict[str, Callable] = {
    "start": lambda *args, **kwargs: thread_start_process(*args, **kwargs),
    "resume": lambda *args, **kwargs: thread_resume_process(*args, **kwargs),
    "validate": thread_validate_workflow,
}


def resume_process(
    process: ProcessTable,
    *,
    user_inputs: Optional[List[State]] = None,
    user: Optional[str] = None,
    broadcast_func: Optional[BroadcastFunc] = None,
) -> UUID:
    """Resume a failed or suspended process.

    Args:
        process: Process from database
        user_inputs: Optional user input from forms
        user: user who resumed this process
        broadcast_func: Optional function to broadcast process data

    Returns:
        process id

    """
    pstat = load_process(process)
    try:
        post_process(pstat.log[0].form, pstat.state.unwrap(), user_inputs=user_inputs or [])
    except FormValidationError:
        logger.exception("Validation errors", user_inputs=user_inputs)
        raise

    resume_func = get_execution_context()["resume"]
    return resume_func(process, user_inputs=user_inputs, user=user, broadcast_func=broadcast_func)


async def _async_resume_processes(
    processes: List[ProcessTable],
    user_name: str,
    broadcast_func: Optional[Callable] = None,
) -> bool:
    """Asynchronously resume multiple failed processes.

    Args:
        processes: Processes from database
        user_name: User who requested resuming the processes

    Returns:
        True if the resume-all operation has been started.
        False if it has not been started because it is already running.

    """
    lock_expiration = max(30, len(processes) // 10)
    if not (lock := await distlock_manager.get_lock("resume-all", lock_expiration)):
        return False

    def run() -> None:
        try:
            for _proc in processes:
                try:
                    process = _get_process(_proc.pid)
                    if process.last_status == ProcessStatus.RUNNING:
                        # Process has been started by something else in the meantime
                        logger.info("Cannot resume a running process", pid=_proc.pid)
                        continue
                    elif process.last_status == ProcessStatus.RESUMED:
                        # Process has been resumed by something else in the meantime
                        logger.info("Cannot resume a resumed process", pid=_proc.pid)
                        continue
                    resume_process(process, user=user_name, broadcast_func=broadcast_func)
                except Exception:
                    logger.exception("Failed to resume process", pid=_proc.pid)
            logger.info("Completed resuming processes")
        finally:
            distlock_manager.release_sync(lock)

    # Start all jobs in the background. BackgroundTasks might be more suited.
    workflow_executor = get_thread_pool()
    process_handle = workflow_executor.submit(run)
    if app_settings.TESTING:
        process_handle.result()
    return True


def abort_process(process: ProcessTable, user: str, broadcast_func: Optional[Callable] = None) -> WFProcess:
    pstat = load_process(process)

    pstat.update(current_user=user)
    return abort_wf(pstat, partial(safe_logstep, broadcast_func=broadcast_func))


def _recoverwf(wf: Workflow, log: List[WFProcess]) -> Tuple[WFProcess, StepList]:
    # Remove all extra steps (Failed, Suspended and Waiting steps add extra steps in db)
    persistent = list(filter(lambda p: not (p.isfailed() or p.issuspend() or p.iswaiting()), log))
    stepcount = len(persistent)

    # Make sure we get the last state from the suspend step (since we removed it before)
    if log and log[-1].issuspend():
        state = log[-1]
    elif persistent:
        state = persistent[-1]
    else:
        state = Success({})

    rest = wf.steps[stepcount:]

    return state, rest


def _restore_log(steps: List[ProcessStepTable]) -> List[WFProcess]:
    result = []
    for step in steps:
        process = WFProcess.from_status(step.status, step.state)

        if not process:
            raise ValueError(step.status)

        result.append(process)
    return result


def load_process(process: ProcessTable) -> ProcessStat:
    workflow = get_workflow(process.workflow)

    if not workflow:
        workflow = removed_workflow

    log = _restore_log(process.steps)
    pstate, remaining = _recoverwf(workflow, log)

    return ProcessStat(
        pid=process.pid,
        workflow=workflow,
        state=pstate,
        log=remaining,
        current_user=SYSTEM_USER,
    )


class ProcessDataBroadcastThread(threading.Thread):
    def __init__(self, _websocket_manager: WebSocketManager, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self.shutdown = False
        self.queue: queue.Queue = queue.Queue()
        self.websocket_manager = _websocket_manager

    def run(self) -> None:
        logger.info("Starting ProcessDataBroadcastThread")
        try:
            loop = asyncio.new_event_loop()  # Create an eventloop specifically for this thread

            while not self.shutdown:
                try:
                    pid, data = self.queue.get(block=True, timeout=1)
                except queue.Empty:
                    continue
                logger.debug(
                    "Threadsafe broadcast data through websocket manager",
                    process_id=pid,
                    where="ProcessDataBroadcastThread",
                    channels=WS_CHANNELS.ALL_PROCESSES,
                )
                loop.run_until_complete(self.websocket_manager.broadcast_data([WS_CHANNELS.ALL_PROCESSES], data))

            loop.close()
            logger.info("Shutdown ProcessDataBroadcastThread")
        except Exception:
            logger.exception("Unhandled exception in ProcessDataBroadcastThread, exiting")

    def stop(self) -> None:
        logger.debug("Sending shutdown signal to ProcessDataBroadcastThread")
        self.shutdown = True
        self.join(timeout=5)
        self.is_alive()


def api_broadcast_process_data(request: Request) -> Optional[BroadcastFunc]:
    """Given a FastAPI request, creates a threadsafe callable for broadcasting process data.

    The callable should be created in API endpoints and provided to start_process,
    resume_process, etc. through the `broadcast_func` param.
    """
    broadcast_queue: queue.Queue = request.app.broadcast_thread.queue

    def _queue_put(pid: UUID, data: Dict) -> None:
        broadcast_queue.put((str(pid), data))

    return _queue_put


class ThreadPoolWorkerStatus(WorkerStatus):
    def __init__(self) -> None:
        super().__init__(executor_type="threadpool")
        thread_pool = get_thread_pool()
        self.number_of_workers_online = getattr(thread_pool, "_max_workers", -1)
        self.number_of_queued_jobs = thread_pool._work_queue.qsize() if hasattr(thread_pool, "_work_queue") else 0
        self.number_of_running_jobs = len(getattr(thread_pool, "_threads", []))



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from collections import defaultdict
from datetime import datetime
from itertools import groupby, zip_longest
from operator import attrgetter
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_type_hints,
)
from uuid import UUID, uuid4

import structlog
from more_itertools import first, flatten, last, one, only
from pydantic import BaseModel, Field, ValidationError
from pydantic.fields import PrivateAttr
from pydantic.main import ModelMetaclass
from pydantic.types import ConstrainedList
from pydantic.typing import get_args, get_origin
from sqlalchemy.orm import selectinload

from orchestrator.db import (
    ProductBlockTable,
    ProductTable,
    SubscriptionInstanceRelationTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionTable,
    db,
)
from orchestrator.domain.lifecycle import (
    ProductLifecycle,
    lookup_specialized_type,
    register_specialized_type,
    validate_lifecycle_status,
)
from orchestrator.services.products import get_product_by_id
from orchestrator.types import (
    SAFE_USED_BY_TRANSITIONS_FOR_STATUS,
    State,
    SubscriptionLifecycle,
    UUIDstr,
    get_possible_product_block_types,
    is_list_type,
    is_of_type,
    is_optional_type,
    is_union_type,
)
from orchestrator.utils.datetime import nowtz
from orchestrator.utils.docs import make_product_block_docstring, make_subscription_model_docstring

logger = structlog.get_logger(__name__)


class ProductNotInRegistryException(Exception):
    pass


class serializable_property(property):
    """Inherit from property class to mark a field in a product block as serializable."""

    pass


def _is_constrained_list_type(type: Type) -> bool:
    """Check if type is a constained list type.

    Example:
        >>> _is_constrained_list_type(List[int])
        False
        >>> class ListType(ConstrainedList):
        ...     min_items = 1
        >>> _is_constrained_list_type(ListType)
        True

    """
    # subclass on typing.List throws exception and there is no good way to test for this
    try:
        is_constrained_list = issubclass(type, ConstrainedList)
    except Exception:
        # Strip generic arguments, it still might be a subclass
        if get_origin(type):
            return _is_constrained_list_type(get_origin(type))  # type: ignore
        else:
            return False

    return is_constrained_list


T = TypeVar("T")  # pragma: no mutate
S = TypeVar("S", bound="SubscriptionModel")  # pragma: no mutate
B = TypeVar("B", bound="ProductBlockModel")  # pragma: no mutate


class DomainModel(BaseModel):
    """Base class for domain models.

    Contains all common Product block/Subscription instance code
    """

    class Config:
        validate_assignment = True  # pragma: no mutate
        validate_all = True  # pragma: no mutate
        arbitrary_types_allowed = True  # pragma: no mutate

    __base_type__: ClassVar[Optional[Type["DomainModel"]]] = None  # pragma: no mutate
    _product_block_fields_: ClassVar[Dict[str, Type]]
    _non_product_block_fields_: ClassVar[Dict[str, Type]]

    def __init_subclass__(
        cls, *args: Any, lifecycle: Optional[List[SubscriptionLifecycle]] = None, **kwargs: Any
    ) -> None:
        super().__init_subclass__()
        cls._find_special_fields()

        if kwargs.keys():
            logger.warning(
                "Unexpected keyword arguments in domain model class",  # pragma: no mutate
                class_name=cls.__name__,
                kwargs=kwargs.keys(),
            )

        # Check if dependency subscription instance models conform to the same lifecycle
        for product_block_field_name, product_block_field_type in cls._get_depends_on_product_block_types().items():
            if lifecycle:
                for lifecycle_status in lifecycle:
                    if is_union_type(
                        product_block_field_type
                    ):  # added to support a list with union of multiple product blocks.
                        product_block_field_type = get_args(product_block_field_type)  # type: ignore

                    if isinstance(product_block_field_type, tuple):
                        for field_type in product_block_field_type:
                            validate_lifecycle_status(product_block_field_name, field_type, lifecycle_status)
                    else:
                        validate_lifecycle_status(product_block_field_name, product_block_field_type, lifecycle_status)

    @classmethod
    def _get_depends_on_product_block_types(
        cls,
    ) -> Dict[str, Union[Type["ProductBlockModel"], Tuple[Type["ProductBlockModel"]]]]:
        """Return all the product block model types.

        This strips any List[..] or Optional[...] types.
        """
        result = {}
        for product_block_field_name, product_block_field_type in cls._product_block_fields_.items():
            if is_union_type(product_block_field_type) and not is_optional_type(product_block_field_type):
                field_type: Union[Type["ProductBlockModel"], Tuple[Type["ProductBlockModel"]]] = get_args(product_block_field_type)  # type: ignore
            # exclude non-Optional Unions as they contain more than one useful element.
            elif is_list_type(product_block_field_type) or (
                is_optional_type(product_block_field_type) and len(get_args(product_block_field_type)) <= 2
            ):
                field_type = first(get_args(product_block_field_type))
            else:
                field_type = product_block_field_type

            result[product_block_field_name] = field_type
        return result

    @classmethod
    def _find_special_fields(cls: Type) -> None:
        """Make and store a list of resource_type fields and product block fields."""
        cls._non_product_block_fields_ = {}
        cls._product_block_fields_ = {}

        if sys.version_info >= (3, 10):
            from inspect import get_annotations
        else:
            # python 3.9 compatibility
            def get_annotations(obj: Any) -> dict:
                if hasattr(obj, "__annotations__"):
                    return obj.__annotations__

                if isinstance(obj, type) and "__annotations__" in obj.__dict__:
                    return obj.__dict__["__annotations__"]

                raise Exception(f"Cannot resolve type annotations for object {obj!r}")

        annotations = get_annotations(cls)

        # Retrieve type hints with evaluated ForwardRefs (for nested blocks)
        type_hints = get_type_hints(cls, localns={cls.__name__: cls})

        # But this also returns inherited fields so cross-check against the annotations
        final_annotations = {k: type_hints[k] for k in annotations}

        for field_name, field_type in final_annotations.items():
            if field_name.startswith("_"):
                continue

            try:
                is_product_block_field = (
                    is_union_type(field_type, DomainModel)
                    or is_list_type(field_type, DomainModel)
                    or is_optional_type(field_type, DomainModel)
                    or is_of_type(field_type, DomainModel)
                )
            except TypeError:
                # issubclass does not work on typing types
                is_product_block_field = False

            # Figure out if this field_name has an alias. Needed sometimes for serializable properties that
            # have a 'real' property with the same name that has a field alias.
            if field := cls.__fields__.get(field_name):
                field_name = field.alias

            # We only want fields that are on this class and not on the related product blocks
            if is_product_block_field:
                cls._product_block_fields_[field_name] = field_type
            else:
                cls._non_product_block_fields_[field_name] = field_type

    @classmethod
    def _init_instances(
        cls, subscription_id: UUID, skip_keys: Optional[List[str]] = None
    ) -> Dict[str, Union[List["ProductBlockModel"], "ProductBlockModel"]]:
        """Initialize default subscription instances.

        When a new domain model is created that is not loaded from an existing subscription.
        We also create all subscription instances for it. This function does that.

        Args:
            skip_keys: list of fields on the class to skip when creating dummy instances.

        Returns:
            A dict with instances to pass to the new model

        """
        if skip_keys is None:
            skip_keys = []

        instances: Dict[str, Union[List[ProductBlockModel], ProductBlockModel]] = {}
        for product_block_field_name, product_block_field_type in cls._product_block_fields_.items():
            if product_block_field_name in skip_keys:
                continue

            if is_list_type(product_block_field_type):
                if _is_constrained_list_type(product_block_field_type):
                    product_block_model = one(get_args(product_block_field_type))
                    if is_union_type(product_block_model):
                        product_block_model = last(get_args(product_block_model))
                    default_value = product_block_field_type()
                    # if constrainedlist has minimum, return that minimum else empty list
                    if product_block_field_type.min_items:
                        logger.debug("creating min_items", type=product_block_field_type)  # pragma: no mutate
                        for _ in range(product_block_field_type.min_items):
                            default_value.append(product_block_model.new(subscription_id=subscription_id))
                else:
                    # a list field of ProductBlockModels without limits gets an empty list
                    default_value = []
            elif is_optional_type(product_block_field_type, ProductBlockModel):
                default_value = None
            elif is_union_type(product_block_field_type):
                raise ValueError(
                    "Union Types must always be `Optional` when calling `.new().` We are unable to detect which type to intialise and Union types always cross subscription boundaries."
                )
            else:
                product_block_model = product_block_field_type
                # Scalar field of a ProductBlockModel expects 1 instance
                default_value = product_block_model.new(subscription_id=subscription_id)
            instances[product_block_field_name] = default_value
        return instances

    @classmethod
    def _load_instances(
        cls,
        db_instances: List[SubscriptionInstanceTable],
        status: SubscriptionLifecycle,
        match_domain_attr: bool = True,
    ) -> Dict[str, Union[Optional["ProductBlockModel"], List["ProductBlockModel"]]]:
        """Load subscription instances for this domain model.

        When a new domain model is loaded from an existing subscription we also load all
        subscription instances for it. This function does that.

        Args:
            db_instances: list of database models to load from
            status: SubscriptionLifecycle of subscription to check if models match
            match_domain_attr: Match domain attribute from relation (not wanted when loading product blocks directly related to subscriptions)

        Returns:
            A dict with instances to pass to the new model

        """

        instances: Dict[str, Union[Optional[ProductBlockModel], List[ProductBlockModel]]] = {}

        def keyfunc(i: SubscriptionInstanceTable) -> str:
            return i.product_block.name

        sorted_instances = sorted(db_instances, key=keyfunc)
        grouped_instances = {k: list(g) for k, g in groupby(sorted_instances, keyfunc)}

        def match_domain_model_attr_if_possible(field_name: str) -> Callable:
            def domain_filter(instance: SubscriptionInstanceTable) -> bool:
                """
                Match domain model attributes.

                This helper is necessary to filter through all relations in a subscription. Not all subscriptions have a
                domain model attribute that is set as it is not always necessary. However when it is set, it is necessary
                to filter through instances depending on that attribute.

                Args:
                    instance: depends on subscription instance

                Returns:
                    Boolean of match.

                """
                # We don't match on the product_blocks directly under subscriptions. They don't have in_use_by relations to those
                if not match_domain_attr:
                    return True

                attr_names = {
                    relation.domain_model_attr
                    for relation in instance.in_use_by_block_relations
                    if relation.domain_model_attr
                }

                # We can assume true is no domain_model_attr is set.
                return not attr_names or field_name in attr_names

            return domain_filter

        for product_block_field_name, product_block_field_type in cls._product_block_fields_.items():
            filter_func = match_domain_model_attr_if_possible(product_block_field_name)

            product_block_model: Any = product_block_field_type
            if is_list_type(product_block_field_type):
                product_block_model = one(get_args(product_block_field_type))

            possible_product_block_types = get_possible_product_block_types(product_block_model)
            field_type_names = list(possible_product_block_types.keys())
            filtered_instances = flatten([grouped_instances.get(name, []) for name in field_type_names])
            instance_list = list(filter(filter_func, filtered_instances))

            if is_list_type(product_block_field_type):
                if product_block_field_name not in grouped_instances:
                    product_block_model_list = []
                    if _is_constrained_list_type(product_block_field_type):
                        product_block_model_list = product_block_field_type()

                product_block_model_list.extend(
                    possible_product_block_types[instance.product_block.name].from_db(
                        subscription_instance=instance, status=status
                    )
                    for instance in instance_list
                )

                instances[product_block_field_name] = product_block_model_list
            else:
                instance = only(instance_list)
                if not is_optional_type(product_block_field_type) and instance is None:
                    raise ValueError("Required subscription instance is missing in database")
                elif is_optional_type(product_block_field_type) and instance is None:
                    instances[product_block_field_name] = None
                elif instance:
                    assert (  # noqa: S101
                        len(possible_product_block_types) is not None
                    ), "Product block model has not been resolved. Unable to continue"
                    instances[product_block_field_name] = possible_product_block_types[
                        instance.product_block.name
                    ].from_db(subscription_instance=instance, status=status)
        return instances

    @classmethod
    def _data_from_lifecycle(cls, other: "DomainModel", status: SubscriptionLifecycle, subscription_id: UUID) -> Dict:
        data = other.dict()

        for field_name, field_type in cls._product_block_fields_.items():
            value = getattr(other, field_name)

            if is_list_type(field_type):
                data[field_name] = []
                list_field_type = one(get_args(field_type))
                possible_product_block_types = get_possible_product_block_types(list_field_type)

                for item in value:
                    data[field_name].append(
                        possible_product_block_types[item.name]._from_other_lifecycle(item, status, subscription_id)
                    )
            else:
                if is_union_type(field_type):
                    if is_optional_type(field_type):
                        data[field_name] = None
                    field_types = get_args(field_type)
                    for f_type in field_types:
                        if value and not isinstance(None, f_type) and f_type.name == value.name:
                            field_type = f_type
                if value:
                    data[field_name] = field_type._from_other_lifecycle(value, status, subscription_id)
        return data

    def _save_instances(
        self, subscription_id: UUID, status: SubscriptionLifecycle
    ) -> Tuple[List[SubscriptionInstanceTable], Dict[str, List[SubscriptionInstanceTable]]]:
        """Save subscription instances for this domain model.

        When a domain model is saved to the database we need to save all depends_on subscription instances for it.

        Args:
            subscription_id: The subscription id
            status: SubscriptionLifecycle of subscription to check if models match

        Returns:
            A list with instances which are saved and a dict with direct depends_on relations.

        """
        saved_instances: List[SubscriptionInstanceTable] = []
        depends_on_instances: Dict[str, List[SubscriptionInstanceTable]] = {}
        for product_block_field, product_block_field_type in self._product_block_fields_.items():
            product_block_models = getattr(self, product_block_field)
            if is_list_type(product_block_field_type):
                field_instance_list = []
                for product_block_model in product_block_models:
                    saved, depends_on_instance = product_block_model.save(
                        subscription_id=subscription_id, status=status
                    )
                    field_instance_list.append(depends_on_instance)
                    saved_instances.extend(saved)
                depends_on_instances[product_block_field] = field_instance_list
            elif (
                is_optional_type(product_block_field_type) or is_union_type(product_block_field_type)
            ) and product_block_models is None:
                pass
            else:
                saved, depends_on_instance = product_block_models.save(subscription_id=subscription_id, status=status)
                depends_on_instances[product_block_field] = [depends_on_instance]
                saved_instances.extend(saved)

        return saved_instances, depends_on_instances

    @classmethod
    def get_properties(cls) -> list[Any]:
        def get_props(klass: type) -> Generator:
            yield from (prop for prop in klass.__dict__ if isinstance(klass.__dict__[prop], serializable_property))

        def get_all_props() -> Generator:
            for klass in cls.mro():
                yield from get_props(klass)

        return list(get_all_props())

    def dict(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Override the dict function to include serializable properties."""
        attribs = super().dict(**kwargs)
        props = self.get_properties()

        # Include and exclude properties
        if include := kwargs.get("include"):
            props = [prop for prop in props if prop in include]
        if exclude := kwargs.get("exclude"):
            props = [prop for prop in props if prop not in exclude]

        # Update the attribute dict with the properties
        if props:
            attribs.update({prop: getattr(self, prop) for prop in props})
        return attribs


class ProductBlockModelMeta(ModelMetaclass):
    """Metaclass used to create product block instances.

    This metaclass is used to make sure the class contains product block metadata.

    This metaclass should not be used directly in the class definition. Instead a new product block model should inherit
    from ProductBlockModel which has this metaclass defined.

    You can find some examples in: :ref:`domain-models`
    """

    __names__: Set[str]
    name: Optional[str]
    product_block_id: UUID
    description: str
    tag: str
    registry: Dict[str, Type["ProductBlockModel"]] = {}  # pragma: no mutate

    def _fix_pb_data(self) -> None:
        if not self.name:
            raise ValueError(f"Cannot create instance of abstract class. Use one of {self.__names__}")

        # Would have been nice to do this in __init_subclass__ but that runs outside the app context so we can't
        # access the db. So now we do it just before we instantiate the instance
        if not hasattr(self, "product_block_id"):
            product_block = ProductBlockTable.query.filter(ProductBlockTable.name == self.name).one()
            self.product_block_id = product_block.product_block_id
            self.description = product_block.description
            self.tag = product_block.tag

    def __call__(self, *args: Any, **kwargs: Any) -> B:  # type: ignore
        self._fix_pb_data()

        kwargs["name"] = self.name

        return super().__call__(*args, **kwargs)


def get_depends_on_product_block_type_list(
    product_block_types: Dict[str, Union[Type["ProductBlockModel"], Tuple[Type["ProductBlockModel"]]]]
) -> List[Type["ProductBlockModel"]]:
    product_blocks_types_in_model = []
    for product_block_type in product_block_types.values():
        if is_union_type(product_block_type):
            for union_product_block_type in get_args(product_block_type):  # type: ignore
                if not isinstance(None, union_product_block_type):
                    product_blocks_types_in_model.append(union_product_block_type)
        else:
            product_blocks_types_in_model.append(product_block_type)

    if product_blocks_types_in_model and isinstance(first(product_blocks_types_in_model), tuple):
        return one(product_blocks_types_in_model)
    return product_blocks_types_in_model


class ProductBlockModel(DomainModel, metaclass=ProductBlockModelMeta):
    r"""Base class for all product block models.

    This class should have been called SubscriptionInstanceModel.

    ProductTable Blocks are represented as dataclasses with pydantic runtime validation.

    Different stages of a subscription lifecycle could require different product block definition. Mainly to support
    mandatory fields when a subscription is active. To support this a lifecycle specific product block definition can
    be created by subclassing the generic product block with keyword argument 'lifecycle' and overriding its fields.

    All product blocks are related to a database ProductBlockTable object through the `product_block_name` that is given
    as class keyword argument.

    Define a product block:
    >>> class BlockInactive(ProductBlockModel, product_block_name="Virtual Circuit"):
    ...    int_field: Optional[int] = None
    ...    str_field: Optional[str] = None

    >>> class Block(BlockInactive, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    ...    int_field: int
    ...    str_field: str

    This example defines a product_block with two different contraints based on lifecycle. `Block` is valid only for `ACTIVE`
    And `BlockInactive` for all other states.
    `product_block_name` must be defined on the base class and need not to be defined on the others

    Create a new empty product block
    >>> example1 = BlockInactive()  # doctest:+SKIP

    Create a new instance based on a dict in the state:
    >>> example2 = BlockInactive(\*\*state)  # doctest:+SKIP

    To retrieve a ProductBlockModel from the database.:
    >>> BlockInactive.from_db(subscription_instance_id)  # doctest:+SKIP
    """

    registry: ClassVar[Dict[str, Type["ProductBlockModel"]]]  # pragma: no mutate
    __names__: ClassVar[Set[str]] = set()
    product_block_id: ClassVar[UUID]
    description: ClassVar[str]
    tag: ClassVar[str]
    _db_model: SubscriptionInstanceTable = PrivateAttr()

    # Product block name. This needs to be an instance var because its part of the API (we expose it to the frontend)
    # Is actually optional since abstract classes dont have it. In practice it is always set
    name: str
    subscription_instance_id: UUID
    owner_subscription_id: UUID
    label: Optional[str] = None

    def __init_subclass__(
        cls,
        *,
        product_block_name: Optional[str] = None,
        lifecycle: Optional[List[SubscriptionLifecycle]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init_subclass__(lifecycle=lifecycle, **kwargs)

        if product_block_name is not None:
            # This is a concrete product block base class (so not a abstract super class or a specific lifecycle version)
            cls.name = product_block_name
            cls.__base_type__ = cls
            cls.__names__ = {cls.name}
            ProductBlockModel.registry[cls.name] = cls
        elif lifecycle is None:
            # Abstract class, no product block name
            cls.name = None  # type:ignore
            cls.__names__ = set()

        # For everything except abstract classes
        if cls.name is not None:
            register_specialized_type(cls, lifecycle)

            # Add ourselves to any super class. That way we can match a superclass to an instance when loading
            for klass in cls.__mro__:
                if issubclass(klass, ProductBlockModel):
                    klass.__names__.add(cls.name)

        cls.__doc__ = make_product_block_docstring(cls, lifecycle)

    @classmethod
    def diff_product_block_in_database(cls) -> Dict[str, Set[str]]:
        """Return any differences between the attrs defined on the domain model and those on product blocks in the database.

        This is only needed to check if the domain model and database models match which would be done during testing...
        """
        if not cls.name:
            # This is a superclass we can't check that
            return {}

        product_block_db: ProductBlockTable = ProductBlockTable.query.filter(
            ProductBlockTable.name == cls.name
        ).one_or_none()
        product_blocks_in_db = {pb.name for pb in product_block_db.depends_on} if product_block_db else set()

        product_blocks_in_model = cls._get_depends_on_product_block_types()
        product_blocks_types_in_model = get_depends_on_product_block_type_list(product_blocks_in_model)

        product_blocks_in_model = set(flatten(map(attrgetter("__names__"), product_blocks_types_in_model)))  # type: ignore

        missing_product_blocks_in_db = product_blocks_in_model - product_blocks_in_db  # type: ignore
        missing_product_blocks_in_model = product_blocks_in_db - product_blocks_in_model  # type: ignore

        resource_types_model = set(cls._non_product_block_fields_)
        resource_types_db = {rt.resource_type for rt in product_block_db.resource_types} if product_block_db else set()

        missing_resource_types_in_db = resource_types_model - resource_types_db
        missing_resource_types_in_model = resource_types_db - resource_types_model

        logger.debug(
            "ProductBlockTable blocks diff",
            product_block_db=product_block_db.name if product_block_db else None,
            product_blocks_in_db=product_blocks_in_db,
            product_blocks_in_model=product_blocks_in_model,
            resource_types_db=resource_types_db,
            resource_types_model=resource_types_model,
            missing_product_blocks_in_db=missing_product_blocks_in_db,
            missing_product_blocks_in_model=missing_product_blocks_in_model,
            missing_resource_types_in_db=missing_resource_types_in_db,
            missing_resource_types_in_model=missing_resource_types_in_model,
        )

        missing_data: Dict[str, Any] = {}
        for product_block_model in product_blocks_types_in_model:
            if product_block_model.name == cls.name or product_block_model.name in missing_data:
                continue
            missing_data.update(product_block_model.diff_product_block_in_database())

        diff: Dict[str, Set[str]] = {
            k: v
            for k, v in {
                "missing_product_blocks_in_db": missing_product_blocks_in_db,
                "missing_product_blocks_in_model": missing_product_blocks_in_model,
                "missing_resource_types_in_db": missing_resource_types_in_db,
                "missing_resource_types_in_model": missing_resource_types_in_model,
            }.items()
            if v
        }

        if diff:
            missing_data[cls.name] = diff

        return missing_data

    @classmethod
    def new(cls: Type[B], subscription_id: UUID, **kwargs: Any) -> B:
        """Create a new empty product block.

        We need to use this instead of the normal constructor because that assumes you pass in
        all required values. That is cumbersome since that means creating a tree of product blocks.

        This is similar to `from_product_id()`
        """
        sub_instances = cls._init_instances(subscription_id, list(kwargs.keys()))

        subscription_instance_id = uuid4()

        # Make sure product block stuff is already set if new is the first usage of this class
        cls._fix_pb_data()

        db_model = SubscriptionInstanceTable(
            product_block_id=cls.product_block_id,
            subscription_instance_id=subscription_instance_id,
            subscription_id=subscription_id,
        )
        db.session.enable_relationship_loading(db_model)
        model = cls(subscription_instance_id=subscription_instance_id, owner_subscription_id=subscription_id, **sub_instances, **kwargs)  # type: ignore
        model._db_model = db_model
        return model

    @classmethod
    def _load_instances_values(cls, instance_values: List[SubscriptionInstanceValueTable]) -> Dict[str, str]:
        """Load non product block fields (instance values).

        Args:
            instance_values: List of instance values from database

        Returns:
            Dict of fields to use for constructor

        """
        instance_values_dict: State = {}
        list_field_names = set()

        # Set default values
        for field_name, field_type in cls._non_product_block_fields_.items():
            # Ensure that empty lists are handled OK
            if is_list_type(field_type):
                instance_values_dict[field_name] = []
                list_field_names.add(field_name)

        for siv in instance_values:
            # check the type of the siv in the instance and act accordingly: only lists and scalar values supported
            resource_type_name = siv.resource_type.resource_type
            if resource_type_name in list_field_names:
                instance_values_dict[resource_type_name].append(siv.value)
            else:
                instance_values_dict[resource_type_name] = siv.value

        # Make sure values are sorted. This already happens when they come from the db.
        # However newly created SubscriptionInstances might not have the correct order
        for field_name in list_field_names:
            instance_values_dict[field_name] = sorted(instance_values_dict[field_name])

        return instance_values_dict

    @classmethod
    def _from_other_lifecycle(
        cls: Type[B],
        other: "ProductBlockModel",
        status: SubscriptionLifecycle,
        subscription_id: UUID,
    ) -> B:
        """Create new domain model from instance while changing the status.

        This makes sure we always have a specific instance.
        """
        if not cls.__base_type__:
            cls = ProductBlockModel.registry.get(other.name, cls)  # type:ignore
            cls = lookup_specialized_type(cls, status)

        data = cls._data_from_lifecycle(other, status, subscription_id)

        model = cls(**data)
        model._db_model = other._db_model
        return model

    @classmethod
    def from_db(
        cls: Type[B],
        subscription_instance_id: Optional[UUID] = None,
        subscription_instance: Optional[SubscriptionInstanceTable] = None,
        status: Optional[SubscriptionLifecycle] = None,
    ) -> B:
        """Create a product block based on a subscription instance from the database.

        This function is similar to `from_subscription()`

        >>> subscription_instance_id = KNOWN_UUID_IN_DB  # doctest:+SKIP
        >>> si_from_db = db.SubscriptionInstanceTable.query.get(subscription_instance_id)  # doctest:+SKIP
        >>> example3 = ProductBlockModel.from_db(subscription_instance=si_from_db)  # doctest:+SKIP
        >>> example4 = ProductBlockModel.from_db(subscription_instance_id=subscription_instance_id)  # doctest:+SKIP
        """
        # Fill values from actual subscription
        if subscription_instance_id:
            subscription_instance = SubscriptionInstanceTable.query.get(subscription_instance_id)
        if subscription_instance:
            subscription_instance_id = subscription_instance.subscription_instance_id
        assert subscription_instance_id  # noqa: S101
        assert subscription_instance  # noqa: S101

        if not status:
            status = SubscriptionLifecycle(subscription_instance.subscription.status)

        if not cls.__base_type__:
            cls = ProductBlockModel.registry.get(subscription_instance.product_block.name, cls)  # type:ignore
            cls = lookup_specialized_type(cls, status)

        elif not issubclass(cls, lookup_specialized_type(cls, status)):
            raise ValueError(f"{cls} is not valid for lifecycle {status}")

        label = subscription_instance.label

        instance_values = cls._load_instances_values(subscription_instance.values)
        sub_instances = cls._load_instances(subscription_instance.depends_on, status)

        try:
            model = cls(
                subscription_instance_id=subscription_instance_id,
                owner_subscription_id=subscription_instance.subscription_id,
                subscription=subscription_instance.subscription,
                label=label,
                **instance_values,  # type: ignore
                **sub_instances,  # type: ignore
            )
            model._db_model = subscription_instance

            return model
        except ValidationError:
            logger.exception(
                "Subscription is not correct in database",
                loaded_instance_values=instance_values,
                loaded_sub_instances=sub_instances,
            )
            raise

    def _save_instance_values(
        self, product_block: ProductBlockTable, current_values: List[SubscriptionInstanceValueTable]
    ) -> List[SubscriptionInstanceValueTable]:
        """Save non product block fields (instance values).

        Returns:
            List of database instances values to save

        """
        resource_types = {rt.resource_type: rt for rt in product_block.resource_types}
        current_values_dict: Dict[str, List[SubscriptionInstanceValueTable]] = defaultdict(list)
        for siv in current_values:
            current_values_dict[siv.resource_type.resource_type].append(siv)

        subscription_instance_values = []
        for field_name, field_type in self._non_product_block_fields_.items():
            assert (  # noqa: S101
                field_name in resource_types
            ), f"Domain model {self.__class__} does not match the ProductBlockTable {product_block.name}, missing: {field_name} {resource_types}"

            resource_type = resource_types[field_name]
            value = getattr(self, field_name)
            if value is None:
                continue
            if is_list_type(field_type):
                for val, siv in zip_longest(value, current_values_dict[field_name]):
                    if val:
                        if siv:
                            siv.value = str(val)
                            subscription_instance_values.append(siv)
                        else:
                            subscription_instance_values.append(
                                SubscriptionInstanceValueTable(resource_type=resource_type, value=str(val))
                            )
            else:
                if field_name in current_values_dict:
                    current_value = current_values_dict[field_name][0]
                    current_value.value = str(value)
                    subscription_instance_values.append(current_value)
                else:
                    subscription_instance_values.append(
                        SubscriptionInstanceValueTable(resource_type=resource_type, value=str(value))
                    )
        return subscription_instance_values

    def _set_instance_domain_model_attrs(
        self,
        subscription_instance: SubscriptionInstanceTable,
        subscription_instance_mapping: Dict[str, List[SubscriptionInstanceTable]],
    ) -> None:
        """
        Save the domain model attribute to the database.

        This function iterates through the subscription instances and stores the domain model attribute in the
        hierarchy relationship.

        Args:
            subscription_instance_mapping: a mapping of the domain model attribute a underlying instances

        Returns:
            None

        """
        depends_on_block_relations = []
        # Set the domain_model_attrs in the database
        for domain_model_attr, instances in subscription_instance_mapping.items():
            instance: SubscriptionInstanceTable
            for index, instance in enumerate(instances):
                relation = SubscriptionInstanceRelationTable(
                    in_use_by_id=subscription_instance.subscription_instance_id,
                    depends_on_id=instance.subscription_instance_id,
                    order_id=index,
                    domain_model_attr=domain_model_attr,
                )
                depends_on_block_relations.append(relation)
        subscription_instance.depends_on_block_relations = depends_on_block_relations

    def save(
        self, *, subscription_id: UUID, status: SubscriptionLifecycle
    ) -> Tuple[List[SubscriptionInstanceTable], SubscriptionInstanceTable]:
        """Save the current model instance to the database.

        This means saving the whole tree of subscription instances and separately saving all instance values for this instance.

        Args:
            status: current SubscriptionLifecycle to check if all constraints match
            subscription_id: Optional subscription id needed if this is a new model

        Returns:
            List of saved instances

        """
        if not self.name:
            raise ValueError(f"Cannot create instance of abstract class. Use one of {self.__names__}")

        # Make sure we have a valid subscription instance database model
        subscription_instance: SubscriptionInstanceTable = SubscriptionInstanceTable.query.get(
            self.subscription_instance_id
        )
        if subscription_instance:
            # Make sure we do not use a mapped session.
            db.session.refresh(subscription_instance)

            # If this is a "foreign" instance we just stop saving and return it so only its relation is saved
            # We should not touch these themselves
            if self.subscription and subscription_instance.subscription_id != subscription_id:
                return [], subscription_instance

            self._db_model = subscription_instance
        else:
            subscription_instance = self._db_model
            # We only need to add to the session if the subscription_instance does not exist.
            db.session.add(subscription_instance)

        subscription_instance.subscription_id = subscription_id

        db.session.flush()

        # Everything is ok, make sure we are of the right class
        specialized_type = lookup_specialized_type(self.__class__, status)
        if specialized_type and not isinstance(self, specialized_type):
            raise ValueError(
                f"Lifecycle status {status} requires specialized type {specialized_type!r}, was: {type(self)!r}"
            )

        # Actually save stuff
        subscription_instance.label = self.label
        subscription_instance.values = self._save_instance_values(
            subscription_instance.product_block, subscription_instance.values
        )

        sub_instances, depends_on_instances = self._save_instances(subscription_id, status)

        # Save the subscription instances relations.
        self._set_instance_domain_model_attrs(subscription_instance, depends_on_instances)

        return sub_instances + [subscription_instance], subscription_instance

    @property
    def subscription(self) -> SubscriptionTable:
        return self.db_model.subscription

    @property
    def db_model(self) -> SubscriptionInstanceTable:
        return self._db_model

    @property
    def in_use_by(self) -> List[SubscriptionInstanceTable]:
        return self._db_model.in_use_by

    @property
    def depends_on(self) -> List[SubscriptionInstanceTable]:
        return self._db_model.depends_on


class ProductModel(BaseModel):
    """Represent the product as defined in the database as a dataclass."""

    class Config:
        validate_assignment = True  # pragma: no mutate
        validate_all = True  # pragma: no mutate
        arbitrary_types_allowed = True  # pragma: no mutate

    product_id: UUID
    name: str
    description: str
    product_type: str
    tag: str
    status: ProductLifecycle
    created_at: Optional[datetime] = None
    end_date: Optional[datetime] = None


class SubscriptionModel(DomainModel):
    r"""Base class for all product subscription models.

    Define a subscription model:
    >>> class SubscriptionInactive(SubscriptionModel, product_type="SP"):  # doctest:+SKIP
    ...    block: Optional[ProductBlockModelInactive] = None

    >>> class Subscription(BlockInactive, lifecycle=[SubscriptionLifecycle.ACTIVE]):  # doctest:+SKIP
    ...    block: ProductBlockModel


    This example defines a subscription model with two different contraints based on lifecycle. `Subscription` is valid only for `ACTIVE`
    And `SubscriptionInactive` for all other states.
    `product_type` must be defined on the base class and need not to be defined on the others

    Create a new empty subscription
    >>> example1 = SubscriptionInactive.from_product_id(product_id, customer_id)  # doctest:+SKIP

    Create a new instance based on a dict in the state:
    >>> example2 = SubscriptionInactive(\*\*state)  # doctest:+SKIP

    To retrieve a ProductBlockModel from the database:
    >>> SubscriptionInactive.from_subscription(subscription_id)  # doctest:+SKIP
    """

    product: ProductModel
    customer_id: UUID
    _db_model: SubscriptionTable = PrivateAttr()
    subscription_id: UUID = Field(default_factory=uuid4)  # pragma: no mutate
    description: str = "Initial subscription"  # pragma: no mutate
    status: SubscriptionLifecycle = SubscriptionLifecycle.INITIAL  # pragma: no mutate
    insync: bool = False  # pragma: no mutate
    start_date: Optional[datetime] = None  # pragma: no mutate
    end_date: Optional[datetime] = None  # pragma: no mutate
    note: Optional[str] = None  # pragma: no mutate

    def __new__(cls, *args: Any, status: Optional[SubscriptionLifecycle] = None, **kwargs: Any) -> "SubscriptionModel":
        # status can be none if created during change_lifecycle
        if status and not issubclass(cls, lookup_specialized_type(cls, status)):
            raise ValueError(f"{cls} is not valid for status {status}")

        return super().__new__(cls)

    def __init_subclass__(
        cls, is_base: bool = False, lifecycle: Optional[List[SubscriptionLifecycle]] = None, **kwargs: Any
    ) -> None:
        super().__init_subclass__(lifecycle=lifecycle, **kwargs)

        if is_base:
            cls.__base_type__ = cls

        if is_base or lifecycle:
            register_specialized_type(cls, lifecycle)

        cls.__doc__ = make_subscription_model_docstring(cls, lifecycle)

    @classmethod
    def diff_product_in_database(cls, product_id: UUID) -> Dict[str, Dict[str, Union[Set[str], Dict[str, Set[str]]]]]:
        """Return any differences between the attrs defined on the domain model and those on product blocks in the database.

        This is only needed to check if the domain model and database models match which would be done during testing...
        """
        product_db = ProductTable.query.get(product_id)
        product_blocks_in_db = {pb.name for pb in product_db.product_blocks} if product_db else set()

        product_blocks_in_model = cls._get_depends_on_product_block_types()
        product_blocks_types_in_model = get_depends_on_product_block_type_list(product_blocks_in_model)

        product_blocks_in_model = set(flatten(map(attrgetter("__names__"), product_blocks_types_in_model)))  # type: ignore

        missing_product_blocks_in_db = product_blocks_in_model - product_blocks_in_db  # type: ignore
        missing_product_blocks_in_model = product_blocks_in_db - product_blocks_in_model  # type: ignore

        fixed_inputs_model = set(cls._non_product_block_fields_)
        fixed_inputs_in_db = {fi.name for fi in product_db.fixed_inputs} if product_db else set()

        missing_fixed_inputs_in_db = fixed_inputs_model - fixed_inputs_in_db
        missing_fixed_inputs_in_model = fixed_inputs_in_db - fixed_inputs_model

        logger.debug(
            "ProductTable blocks diff",
            product_block_db=product_db.name if product_db else None,
            product_blocks_in_db=product_blocks_in_db,
            product_blocks_in_model=product_blocks_in_model,
            fixed_inputs_in_db=fixed_inputs_in_db,
            fixed_inputs_model=fixed_inputs_model,
            missing_product_blocks_in_db=missing_product_blocks_in_db,
            missing_product_blocks_in_model=missing_product_blocks_in_model,
            missing_fixed_inputs_in_db=missing_fixed_inputs_in_db,
            missing_fixed_inputs_in_model=missing_fixed_inputs_in_model,
        )

        missing_data_depends_on_blocks: Dict[str, Set[str]] = {}
        for product_block_in_model in product_blocks_types_in_model:
            missing_data_depends_on_blocks.update(product_block_in_model.diff_product_block_in_database())

        diff: Dict[str, Union[Set[str], Dict[str, Set[str]]]] = {
            k: v
            for k, v in {
                "missing_product_blocks_in_db": missing_product_blocks_in_db,
                "missing_product_blocks_in_model": missing_product_blocks_in_model,
                "missing_fixed_inputs_in_db": missing_fixed_inputs_in_db,
                "missing_fixed_inputs_in_model": missing_fixed_inputs_in_model,
                "missing_in_depends_on_blocks": missing_data_depends_on_blocks,
            }.items()
            if v
        }

        missing_data: Dict[str, Dict[str, Union[Set[str], Dict[str, Set[str]]]]] = {}
        if diff:
            missing_data[product_db.name] = diff

        return missing_data

    @classmethod
    def from_product_id(
        cls: Type[S],
        product_id: Union[UUID, UUIDstr],
        customer_id: Union[UUID, UUIDstr],
        status: SubscriptionLifecycle = SubscriptionLifecycle.INITIAL,
        description: Optional[str] = None,
        insync: bool = False,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        note: Optional[str] = None,
    ) -> S:
        """Use product_id (and customer_id) to return required fields of a new empty subscription."""
        # Caller wants a new instance and provided a product_id and customer_id
        product_db = ProductTable.query.get(product_id)
        product = ProductModel(
            product_id=product_db.product_id,
            name=product_db.name,
            description=product_db.description,
            product_type=product_db.product_type,
            tag=product_db.tag,
            status=product_db.status,
            created_at=product_db.created_at,
            end_date=product_db.end_date,
        )

        if description is None:
            description = f"Initial subscription of {product.description}"

        subscription_id = uuid4()
        subscription = SubscriptionTable(
            subscription_id=subscription_id,
            product_id=product_id,
            customer_id=customer_id,
            description=description,
            status=status.value,
            insync=insync,
            start_date=start_date,
            end_date=end_date,
            note=note,
        )
        db.session.add(subscription)

        fixed_inputs = {fi.name: fi.value for fi in product_db.fixed_inputs}
        instances = cls._init_instances(subscription_id)

        if isinstance(customer_id, str):
            customer_id = UUID(customer_id)

        model = cls(
            product=product,
            customer_id=customer_id,
            subscription_id=subscription_id,
            description=description,
            status=status,
            insync=insync,
            start_date=start_date,
            end_date=end_date,
            note=note,
            **fixed_inputs,
            **instances,
        )
        model._db_model = subscription
        return model

    @classmethod
    def from_other_lifecycle(
        cls: Type[S],
        other: "SubscriptionModel",
        status: SubscriptionLifecycle,
    ) -> S:
        """Create new domain model from instance while changing the status.

        This makes sure we always have a specific instance.
        """
        if not cls.__base_type__:
            # Import here to prevent cyclic imports
            from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY

            cls = SUBSCRIPTION_MODEL_REGISTRY.get(other.product.name, cls)  # type:ignore
            cls = lookup_specialized_type(cls, status)

        # this will raise ValueError when wrong lifecycle transitions are detected in the new domain model
        validate_lifecycle_change(other, status)

        data = cls._data_from_lifecycle(other, status, other.subscription_id)
        data["status"] = status
        if data["start_date"] is None and status == SubscriptionLifecycle.ACTIVE:
            data["start_date"] = nowtz()
        if data["end_date"] is None and status == SubscriptionLifecycle.TERMINATED:
            data["end_date"] = nowtz()

        model = cls(**data)
        model._db_model = other._db_model

        return model

    # Some common functions shared by from_other_product and from_subscription
    @classmethod
    def _get_subscription(cls: Type[S], subscription_id: Union[UUID, UUIDstr]) -> Any:
        return SubscriptionTable.query.options(
            selectinload(SubscriptionTable.instances)
            .selectinload(SubscriptionInstanceTable.product_block)
            .selectinload(ProductBlockTable.resource_types),
            selectinload(SubscriptionTable.instances).selectinload(SubscriptionInstanceTable.in_use_by_block_relations),
            selectinload(SubscriptionTable.instances).selectinload(SubscriptionInstanceTable.values),
        ).get(subscription_id)

    @classmethod
    def _to_product_model(cls: Type[S], product: ProductTable) -> ProductModel:
        return ProductModel(
            product_id=product.product_id,
            name=product.name,
            description=product.description,
            product_type=product.product_type,
            tag=product.tag,
            status=product.status,
            created_at=product.created_at if product.created_at else None,
            end_date=product.end_date if product.end_date else None,
        )

    @classmethod
    def from_other_product(
        cls: Type[S],
        old_instantiation: S,
        new_product_id: Union[UUID, str],
        new_root: Optional[Tuple[str, ProductBlockModel]] = None,
    ) -> S:
        db_product = get_product_by_id(new_product_id)

        subscription = cls._get_subscription(old_instantiation.subscription_id)
        product = cls._to_product_model(db_product)

        status = SubscriptionLifecycle(subscription.status)

        if not cls.__base_type__:
            # Import here to prevent cyclic imports
            from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY

            cls = SUBSCRIPTION_MODEL_REGISTRY.get(subscription.product.name, cls)  # type:ignore
            cls = lookup_specialized_type(cls, status)
        elif not issubclass(cls, lookup_specialized_type(cls, status)):
            raise ValueError(f"{cls} is not valid for lifecycle {status}")

        fixed_inputs = {fi.name: fi.value for fi in db_product.fixed_inputs}

        if new_root:
            name, product_block = new_root
            instances = {name: product_block}
        else:
            instances = cls._load_instances(subscription.instances, status, match_domain_attr=False)  # type:ignore

        try:
            model = cls(
                product=product,
                customer_id=subscription.customer_id,
                subscription_id=subscription.subscription_id,
                description=subscription.description,
                status=status,
                insync=subscription.insync,
                start_date=subscription.start_date,
                end_date=subscription.end_date,
                note=subscription.note,
                **fixed_inputs,
                **instances,
            )
            model._db_model = subscription
            return model
        except ValidationError:
            logger.exception(
                "Subscription is not correct in database", loaded_fixed_inputs=fixed_inputs, loaded_instances=instances
            )
            raise

    @classmethod
    def from_subscription(cls: Type[S], subscription_id: Union[UUID, UUIDstr]) -> S:
        """Use a subscription_id to return required fields of an existing subscription."""
        subscription = cls._get_subscription(subscription_id)
        if subscription is None:
            raise ValueError(f"Subscription with id: {subscription_id}, does not exist")
        product = cls._to_product_model(subscription.product)

        status = SubscriptionLifecycle(subscription.status)

        if not cls.__base_type__:
            # Import here to prevent cyclic imports
            from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY

            try:
                cls = SUBSCRIPTION_MODEL_REGISTRY[subscription.product.name]  # type:ignore
            except KeyError:
                raise ProductNotInRegistryException(
                    f"'{subscription.product.name}' is not found within the SUBSCRIPTION_MODEL_REGISTRY"
                )
            cls = lookup_specialized_type(cls, status)
        elif not issubclass(cls, lookup_specialized_type(cls, status)):
            raise ValueError(f"{cls} is not valid for lifecycle {status}")

        fixed_inputs = {fi.name: fi.value for fi in subscription.product.fixed_inputs}

        instances = cls._load_instances(subscription.instances, status, match_domain_attr=False)

        try:
            model = cls(
                product=product,
                customer_id=subscription.customer_id,
                subscription_id=subscription.subscription_id,
                description=subscription.description,
                status=status,
                insync=subscription.insync,
                start_date=subscription.start_date,
                end_date=subscription.end_date,
                note=subscription.note,
                **fixed_inputs,
                **instances,
            )
            model._db_model = subscription
            return model
        except ValidationError:
            logger.exception(
                "Subscription is not correct in database", loaded_fixed_inputs=fixed_inputs, loaded_instances=instances
            )
            raise

    def save(self) -> None:
        """Save the subscription to the database."""
        specialized_type = lookup_specialized_type(self.__class__, self.status)
        if specialized_type and not isinstance(self, specialized_type):
            raise ValueError(
                f"Lifecycle status {self.status.value} requires specialized type {specialized_type!r}, was: {type(self)!r}"
            )

        sub = SubscriptionTable.query.options(
            selectinload(SubscriptionTable.instances)
            .selectinload(SubscriptionInstanceTable.product_block)
            .selectinload(ProductBlockTable.resource_types),
            selectinload(SubscriptionTable.instances).selectinload(SubscriptionInstanceTable.values),
        ).get(self.subscription_id)
        if not sub:
            sub = self._db_model

        # Make sure we refresh the object and not use an already mapped object
        db.session.refresh(sub)

        self._db_model = sub
        sub.product_id = self.product.product_id
        sub.customer_id = self.customer_id
        sub.description = self.description
        sub.status = self.status.value
        sub.insync = self.insync
        sub.start_date = self.start_date
        sub.end_date = self.end_date
        sub.note = self.note

        db.session.add(sub)
        db.session.flush()  # Sends INSERT and returns subscription_id without committing transaction

        old_instances_dict = {instance.subscription_instance_id: instance for instance in sub.instances}

        saved_instances, depends_on_instances = self._save_instances(self.subscription_id, self.status)

        for instances in depends_on_instances.values():
            for instance in instances:
                if instance.subscription_id != self.subscription_id:
                    raise ValueError(
                        "Attempting to save a Foreign `Subscription Instance` directly below a subscription. "
                        "This is not allowed."
                    )
        sub.instances = saved_instances

        # Calculate what to remove
        instances_set = {instance.subscription_instance_id for instance in sub.instances}
        for instance_id in instances_set:
            old_instances_dict.pop(instance_id, None)

        # What's left should be removed
        for instance in old_instances_dict.values():
            db.session.delete(instance)

        db.session.flush()

    @property
    def db_model(self) -> SubscriptionTable:
        return self._db_model


def _validate_lifecycle_change_for_product_block(
    used_by: SubscriptionInstanceTable,
    product_block_model: ProductBlockModel,
    status: SubscriptionLifecycle,
    description: str,
) -> None:
    """Validate if a lifecycle change for a single product model is possible."""

    logger.debug(
        "Checking the parent relations",
        parent_status=used_by.subscription.status,
        parent_description=used_by.subscription.description,
        self_status=status,
        self_description=description,
    )
    if (
        used_by.subscription != product_block_model.subscription
        and used_by.subscription.status not in SAFE_USED_BY_TRANSITIONS_FOR_STATUS[status]
    ):
        raise ValueError(
            f"Unsafe status change of Subscription with depending subscriptions: "
            f"{list(map(lambda instance: description, product_block_model.in_use_by))}"
        )


def validate_lifecycle_change(
    other: "SubscriptionModel",
    status: SubscriptionLifecycle,
) -> None:
    """Validate if a lifecycle change for a subscription model is possible.

    It will traverse all product blocks and check the `in_use_by` status to ensure that the lifecycle change
    is allowed.

    Note: A `ValueError` will be raised when a unsafe status change is found
    """
    for product_block_field, product_block_field_type in other._product_block_fields_.items():
        product_block_models = getattr(other, product_block_field)
        if is_list_type(product_block_field_type):
            for product_block_model in product_block_models:
                used_by_generator = (used_by for used_by in product_block_model.in_use_by if used_by)
                for used_by in used_by_generator:
                    _validate_lifecycle_change_for_product_block(
                        used_by, product_block_model, status, other.description
                    )
        elif (
            is_optional_type(product_block_field_type) or is_union_type(product_block_field_type)
        ) and product_block_models is None:
            pass
        else:
            used_by_generator = (used_by for used_by in product_block_models.in_use_by if used_by)
            for used_by in used_by_generator:
                _validate_lifecycle_change_for_product_block(used_by, product_block_models, status, other.description)

    logger.info(
        "Lifecycle validation check ok",
        subscription_id=other.subscription_id,
        subscription_description=other.description,
        status=status,
    )


SI = TypeVar("SI", covariant=True)  # pragma: no mutate


class SubscriptionInstanceList(ConstrainedList, List[SI]):  # type: ignore
    """Shorthand to create constrained lists of product blocks."""

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)

        # Copy generic argument (SI) if not set explicitly
        # This makes a lot of assumptions about the internals of `typing`
        if "__orig_bases__" in cls.__dict__ and cls.__dict__["__orig_bases__"]:
            generic_base_cls = cls.__dict__["__orig_bases__"][0]
            if not hasattr(generic_base_cls, "item_type") and get_args(generic_base_cls):
                cls.item_type = get_args(generic_base_cls)[0]

        # Make sure __args__ is set
        cls.__args__ = (cls.item_type,)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from functools import partial, wraps
from pathlib import Path
from typing import Any, Optional

import structlog
from jinja2 import Environment

from orchestrator.cli.generator.generator.helpers import (
    get_product_file_name,
    get_workflow,
    product_types_module,
    root_product_block,
)
from orchestrator.cli.generator.generator.settings import product_generator_settings
from orchestrator.cli.generator.generator.translations import add_workflow_translations
from orchestrator.cli.generator.generator.validations import get_validations, get_validations_for_modify

logger = structlog.getLogger(__name__)


def find_field_with_name(config: dict, field_name: str) -> dict:
    product_blocks = config.get("product_blocks", [])
    for pb in product_blocks:
        for field in pb.get("fields", []):
            if field["name"] == field_name:
                return {f"{field_name}_path": pb["name"]}
    return {}


def add_optional_ims_config(config: dict) -> dict:
    if found := find_field_with_name(config, "ims_circuit_id"):
        return config | found
    else:
        return config


def add_optional_nso_config(config: dict) -> dict:
    if found := find_field_with_name(config, "nso_service_id"):
        return config | found
    else:
        return config


def insert_lazy_workflow_instances(environment: Environment, config: dict, writer: Callable) -> None:
    template = environment.get_template("lazy_workflow_instance.j2")
    content = template.render(product=config)

    path = f"{product_generator_settings.WORKFLOWS_PATH}/__init__.py"
    writer(path, content, append=True)


def generate_workflows(context: dict) -> None:
    config = context["config"]
    environment = context["environment"]
    writer = context["writer"]

    create_workflow_paths(config)

    config = add_optional_nso_config(config)
    config = add_optional_ims_config(config)

    generate_shared_workflow_files(environment, config, writer)
    generate_create_workflow(environment, config, writer)
    generate_modify_workflow(environment, config, writer)
    generate_validate_workflow(environment, config, writer)
    generate_terminate_workflow(environment, config, writer)

    add_workflow_translations(config, writer)

    # insert_lazy_workflow_instances(environment, config, writer)


def workflow_folder(config: dict) -> str:
    folder = get_product_file_name(config)
    return f"{product_generator_settings.WORKFLOWS_PATH}/{folder}"


def shared_workflow_folder(config: dict) -> str:
    return f"{workflow_folder(config)}/shared"


def create_workflow_paths(config: dict) -> None:
    path = f"{workflow_folder(config)}/shared"
    Path(path).mkdir(parents=True, exist_ok=True)


def get_workflow_path(config: dict, workflow_type: str) -> str:
    file_name = get_product_file_name(config)
    return f"{workflow_folder(config)}/{workflow_type}_{file_name}.py"


def generate_shared_workflow_files(environment: Environment, config: dict, writer: Callable) -> None:
    product_block = root_product_block(config)
    validations, _ = get_validations(config)

    template = environment.get_template("shared_forms.j2")
    content = template.render(product=config, product_block=product_block, validations=validations)

    path = f"{shared_workflow_folder(config)}/forms.py"

    writer(path, content)


def generate_workflow(f: Optional[Callable] = None, workflow: Optional[str] = None) -> Callable:
    if f is None:
        return partial(generate_workflow, workflow=workflow)

    @wraps(f)
    def wrapper(environment: Environment, config: dict, writer: Callable) -> Any:
        def workflow_enabled() -> bool:
            return all(wf.get("enabled", True) for wf in config.get("workflows", []) if wf["name"] == workflow)

        if workflow_enabled():
            return f(environment, config, writer)

    return wrapper


@generate_workflow(workflow="create")
def generate_create_workflow(environment: Environment, config: dict, writer: Callable) -> None:
    product_block = root_product_block(config)
    validations, validation_imports = get_validations(config)

    template = environment.get_template("create_product.j2")
    content = template.render(
        product=config, product_block=product_block, validations=validations, validation_imports=validation_imports
    )

    path = get_workflow_path(config, "create")

    writer(path, content)


@generate_workflow(workflow="modify")
def generate_modify_workflow(environment: Environment, config: dict, writer: Callable) -> None:
    product_block = root_product_block(config)
    validations, validation_imports = get_validations_for_modify(config)

    template = environment.get_template("modify_product.j2")
    content = template.render(
        product=config, product_block=product_block, validations=validations, validation_imports=validation_imports
    )

    path = get_workflow_path(config, "modify")

    writer(path, content)


@generate_workflow(workflow="validate")
def generate_validate_workflow(environment: Environment, config: dict, writer: Callable) -> None:
    workflow = get_workflow(config, "validate")
    validations = workflow.get("validations", [])

    template = environment.get_template("validate_product.j2")
    content = template.render(product=config, validations=validations, product_types_module=product_types_module)

    path = get_workflow_path(config, "validate")

    writer(path, content)


@generate_workflow(workflow="terminate")
def generate_terminate_workflow(environment: Environment, config: dict, writer: Callable) -> None:
    workflow = get_workflow(config, "terminate")
    validations = workflow.get("validations", [])

    template = environment.get_template("terminate_product.j2")
    content = template.render(product=config, validations=validations, product_types_module=product_types_module)

    path = get_workflow_path(config, "terminate")

    writer(path, content)



import json
from http import HTTPStatus
from typing import Union

from fastapi import Response


def get_product_query(
    first: int = 10,
    after: int = 0,
    filter_by: Union[list[str], None] = None,
    sort_by: Union[list[dict[str, str]], None] = None,
) -> bytes:
    query = """
query ProductQuery($first: Int!, $after: Int!, $filterBy: [GraphqlFilter!], $sortBy: [GraphqlSort!]) {
  products(first: $first, after: $after, filterBy: $filterBy, sortBy: $sortBy) {
    page {
      createdAt
      description
      endDate
      fixedInputs {
        createdAt
        fixedInputId
        name
        productId
        value
      }
      name
      productBlocks {
        createdAt
        description
        endDate
        name
        productBlockId
        resourceTypes {
          description
          resourceType
          resourceTypeId
        }
        status
        tag
      }
      productId
      productType
      status
      tag
      workflows {
        createdAt
        description
        target
        name
        workflowId
      }
    }
    pageInfo {
      endCursor
      hasNextPage
      hasPreviousPage
      startCursor
      totalItems
    }
  }
}
    """
    return json.dumps(
        {
            "operationName": "ProductQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
            },
        }
    ).encode("utf-8")


def get_products_with_related_subscriptions_query(
    first: int = 10,
    after: int = 0,
    filter_by: Union[list[str], None] = None,
    sort_by: Union[list[dict[str, str]], None] = None,
) -> bytes:
    query = """
query ProductQuery($first: Int!, $after: Int!, $filterBy: [GraphqlFilter!], $sortBy: [GraphqlSort!]) {
  products(first: $first, after: $after, filterBy: $filterBy, sortBy: $sortBy) {
    page {
      createdAt
      description
      endDate
      fixedInputs {
        createdAt
        fixedInputId
        name
        productId
        value
      }
      name
      productBlocks {
        createdAt
        description
        endDate
        name
        productBlockId
        resourceTypes {
          description
          resourceType
          resourceTypeId
        }
        status
        tag
      }
      productId
      productType
      status
      tag
      workflows {
        createdAt
        description
        target
        name
        workflowId
      }
      subscriptions {
        page {
          description
          subscriptionId
          productId
          status
          insync
          note
          startDate
          endDate
        }
      }
    }
    pageInfo {
      endCursor
      hasNextPage
      hasPreviousPage
      startCursor
      totalItems
    }
  }
}
    """
    return json.dumps(
        {
            "operationName": "ProductQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
            },
        }
    ).encode("utf-8")


def test_product_query(test_client, generic_product_1, generic_product_2, generic_product_3):
    data = get_product_query(first=2)
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    products_data = result["data"]["products"]
    products = products_data["page"]
    pageinfo = products_data["pageInfo"]

    assert len(products) == 2

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 1,
        "totalItems": "3",
    }


def test_product_has_previous_page(test_client, generic_product_1, generic_product_2, generic_product_3):
    data = get_product_query(after=1)
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    products_data = result["data"]["products"]
    products = products_data["page"]
    pageinfo = products_data["pageInfo"]

    assert pageinfo == {
        "endCursor": 2,
        "hasNextPage": False,
        "hasPreviousPage": True,
        "startCursor": 1,
        "totalItems": "3",
    }

    assert len(products) == 2
    assert products[0]["name"] == "Product 2"
    assert products[1]["name"] == "Product 3"


def test_products_filter_by_product_block(test_client, generic_product_1, generic_product_2, generic_product_3):
    data = get_product_query(
        filter_by=[{"field": "product_blocks", "value": "PB_1-PB_3"}],
        sort_by=[{"field": "name", "order": "ASC"}],
    )
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    products_data = result["data"]["products"]
    products = products_data["page"]
    pageinfo = products_data["pageInfo"]

    assert pageinfo == {
        "endCursor": 1,
        "hasNextPage": False,
        "hasPreviousPage": False,
        "startCursor": 0,
        "totalItems": "2",
    }
    assert products[0]["name"] == "Product 1"
    assert products[1]["name"] == "Product 2"


def test_products_sort_by_tag(test_client, generic_product_1, generic_product_2, generic_product_3):
    data = get_product_query(sort_by=[{"field": "tag", "order": "DESC"}])
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    assert HTTPStatus.OK == response.status_code
    result = response.json()

    products_data = result["data"]["products"]
    products = products_data["page"]
    pageinfo = products_data["pageInfo"]

    assert [prod["tag"] for prod in products] == ["GEN3", "GEN2", "GEN1"]
    assert pageinfo == {
        "endCursor": 2,
        "hasNextPage": False,
        "hasPreviousPage": False,
        "startCursor": 0,
        "totalItems": "3",
    }


def test_single_product_with_subscriptions(
    test_client, mocked_processes, generic_product_1, generic_subscription_2, generic_subscription_1
):
    product_id = str(generic_product_1.product_id)
    # when

    data = get_products_with_related_subscriptions_query(filter_by=[{"field": "product_id", "value": product_id}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    products_data = result["data"]["products"]
    products = products_data["page"]
    pageinfo = products_data["pageInfo"]

    assert len(products) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": "1",
    }
    assert products[0]["productId"] == product_id
    assert products[0]["subscriptions"]["page"][0]["subscriptionId"] == generic_subscription_1



from http import HTTPStatus

from orchestrator.utils.errors import ApiException, ProcessFailure, error_state_to_dict


class RESTResponse:  # From openapi-generator generated clients
    def __init__(self, status, reason, data, headers):
        self.status = status
        self.reason = reason
        self.data = data
        self.headers = headers

    def getheaders(self):
        """Return a dictionary of the response headers."""
        return self.headers


def test_error_state_to_dict():
    e = Exception("bla")
    assert error_state_to_dict(e) == {
        "class": "Exception",
        "error": "bla",
        "traceback": "Exception: bla\n",
    }

    e = ApiException(status=HTTPStatus.NOT_FOUND, reason="Not Found")
    assert error_state_to_dict(e) == {
        "body": None,
        "class": "ApiException",
        "error": "Not Found",
        "headers": "",
        "status_code": HTTPStatus.NOT_FOUND,
        "traceback": "ApiException: (404)\n" "Reason: Not Found\n" "\n",
    }

    e = ApiException(
        http_resp=RESTResponse(HTTPStatus.NOT_FOUND, "Not Found", "Body", {"Header": "value", "Content-type": "bogus"})
    )
    assert error_state_to_dict(e) == {
        "body": "Body",
        "class": "ApiException",
        "error": "Not Found",
        "headers": "Header: value\nContent-type: bogus",
        "status_code": HTTPStatus.NOT_FOUND,
        "traceback": "ApiException: (404)\nReason: Not Found\nHTTP response headers: {'Header': 'value', 'Content-type': 'bogus'}\nHTTP response body: Body\n\n",
    }

    e = ProcessFailure(message="Something went wrong", details={"foo": "bar"})
    assert error_state_to_dict(e) == {
        "class": "ProcessFailure",
        "details": {"foo": "bar"},
        "error": "Something went wrong",
        "traceback": "ProcessFailure: ('Something went wrong', {'foo': 'bar'})\n",
    }
